"""
Utility functions to help with URL path lookups using function names for our FastAPI
servers. This makes reference to the route_manifest.yaml file that is also saved in
this directory. This routes_manifest.yaml file should be regenerated whenver changes
are made to the API endpoints. This can be done using the 'generate_route_manifest'
CLI function.
"""

from __future__ import annotations

import re
from functools import lru_cache
from logging import getLogger
from pathlib import Path
from typing import Any

import yaml

import murfey.util

logger = getLogger("murfey.util.api")

route_manifest_file = Path(murfey.util.__path__[0]) / "route_manifest.yaml"


@lru_cache(maxsize=1)  # Load the manifest once and reuse
def load_route_manifest(
    file: Path = route_manifest_file,
):
    with open(file, "r") as f:
        return yaml.safe_load(f)


def find_unique_index(
    pattern: str,
    candidates: list[str],
    exact: bool = False,  # Allows partial matches
) -> int:
    """
    Finds the index of a unique entry in a list.
    """
    counter = 0
    matches = []
    index = 0
    for i, candidate in enumerate(candidates):
        if (not exact and pattern in candidate) or (exact and pattern == candidate):
            counter += 1
            matches.append(candidate)
            index = i
    if counter == 0:
        message = f"No match found for {pattern!r}"
        logger.error(message)
        raise KeyError(message)
    if counter > 1:
        message = f"Ambiguous match for {pattern!r}: {matches}"
        logger.error(message)
        raise KeyError(message)
    return index


def render_path(path_template: str, kwargs: dict) -> str:
    """
    Replace all FastAPI-style {param[:converter]} path parameters with corresponding
    values from kwargs.
    """

    pattern = re.compile(r"{([^}]+)}")  # Look for all path params

    def replace(match):
        raw_str = match.group(1)
        param_name = raw_str.split(":")[0]  # Ignore :converter in the field
        if param_name not in kwargs:
            message = (
                f"Error constructing URL for {path_template!r}; "
                f"missing path parameter {param_name!r}"
            )
            logger.error(message)
            raise KeyError(message)
        return str(kwargs[param_name])

    return pattern.sub(replace, path_template)


def url_path_for(
    router_name: str,  # With logic for partial matches
    function_name: str,  # With logic for partial matches
    **kwargs,  # Takes any path param and matches it against curly bracket contents
):
    """
    Utility function that takes the function name and API router name, along with all
    necessary path parameters, retrieves the matching URL path template from the route
    manifest, and returns a correctly populated instance of the URL path.
    """
    # Use 'Any' first and slowly reveal types as it is unpacked
    route_manifest: dict[str, list[Any]] = load_route_manifest()

    # Load the routes in the desired router
    routers = list(route_manifest.keys())
    routes: list[dict[str, Any]] = route_manifest[
        routers[find_unique_index(router_name, routers, exact=False)]
    ]

    # Search router for the function
    route_info = routes[
        find_unique_index(function_name, [r["function"] for r in routes], exact=True)
    ]

    # Unpack the dictionary
    route_path: str = route_info["path"]
    path_params: list[dict[str, str]] = route_info["path_params"]

    # Validate the stored path params against the ones provided
    if path_params:
        for path_param in path_params:
            param_name = path_param["name"]
            param_type = path_param["type"]
            if param_name not in kwargs.keys():
                message = (
                    f"Error validating parameters for {function_name!r}; "
                    f"path parameter {param_name!r} was not provided"
                )
                logger.error(message)
                raise KeyError(message)
            # Skip complicated type resolution for now
            if param_type.startswith("typing."):
                continue
            elif type(kwargs[param_name]).__name__ not in param_type:
                message = (
                    f"Error validating parameters for {function_name!r}; "
                    f"{param_name!r} must be {param_type!r}, "
                    f"received {type(kwargs[param_name]).__name__!r}"
                )
                logger.error(message)
                raise TypeError(message)

    # Render and return the path
    return render_path(route_path, kwargs)


if __name__ == "__main__":
    # Run test on some existing routes
    url_path = url_path_for(
        "workflow.tomo_router",
        "register_tilt",
        visit_name="nt15587-15",
        session_id=2,
    )
    print(url_path)



from __future__ import annotations

import logging

from backports.entry_points_selectable import entry_points
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from prometheus_client import make_asgi_app
from pydantic_settings import BaseSettings

import murfey.server
import murfey.server.api.auth
import murfey.server.api.bootstrap
import murfey.server.api.clem
import murfey.server.api.display
import murfey.server.api.file_io_frontend
import murfey.server.api.file_io_instrument
import murfey.server.api.hub
import murfey.server.api.instrument
import murfey.server.api.mag_table
import murfey.server.api.processing_parameters
import murfey.server.api.prometheus
import murfey.server.api.session_control
import murfey.server.api.session_info
import murfey.server.api.websocket
import murfey.server.api.workflow
from murfey.server import template_files
from murfey.util.config import get_security_config

log = logging.getLogger("murfey.server.main")

tags_metadata = [murfey.server.api.bootstrap.tag]


class Settings(BaseSettings):
    murfey_machine_configuration: str = ""


security_config = get_security_config()

settings = Settings()

app = FastAPI(title="Murfey server", debug=True, openapi_tags=tags_metadata)

metrics_app = make_asgi_app()
app.mount("/metrics", metrics_app)

app.add_middleware(
    CORSMiddleware,
    allow_origins=security_config.allow_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.mount("/static", StaticFiles(directory=template_files / "static"), name="static")
app.mount("/images", StaticFiles(directory=template_files / "images"), name="images")

# Add router endpoints to the API
app.include_router(murfey.server.api.bootstrap.version)
app.include_router(murfey.server.api.bootstrap.bootstrap)
app.include_router(murfey.server.api.bootstrap.cygwin)
app.include_router(murfey.server.api.bootstrap.msys2)
app.include_router(murfey.server.api.bootstrap.rust)
app.include_router(murfey.server.api.bootstrap.pypi)
app.include_router(murfey.server.api.bootstrap.plugins)

app.include_router(murfey.server.api.auth.router)

app.include_router(murfey.server.api.hub.router)
app.include_router(murfey.server.api.display.router)
app.include_router(murfey.server.api.processing_parameters.router)

app.include_router(murfey.server.api.file_io_frontend.router)
app.include_router(murfey.server.api.file_io_instrument.router)

app.include_router(murfey.server.api.instrument.router)

app.include_router(murfey.server.api.mag_table.router)

app.include_router(murfey.server.api.session_control.router)
app.include_router(murfey.server.api.session_control.spa_router)
app.include_router(murfey.server.api.session_control.tomo_router)

app.include_router(murfey.server.api.session_info.router)
app.include_router(murfey.server.api.session_info.correlative_router)
app.include_router(murfey.server.api.session_info.spa_router)
app.include_router(murfey.server.api.session_info.tomo_router)

app.include_router(murfey.server.api.workflow.router)
app.include_router(murfey.server.api.workflow.correlative_router)
app.include_router(murfey.server.api.workflow.spa_router)
app.include_router(murfey.server.api.workflow.tomo_router)
app.include_router(murfey.server.api.clem.router)

app.include_router(murfey.server.api.prometheus.router)

app.include_router(murfey.server.api.websocket.ws)

# Search external packages for additional routers to include in Murfey
for r in entry_points(group="murfey.routers"):
    app.include_router(r.load())



from __future__ import annotations

import argparse
import configparser
import logging
import os
import platform
import shutil
import sys
import time
import webbrowser
from pathlib import Path
from pprint import pprint
from queue import Queue
from typing import Literal
from urllib.parse import ParseResult, urlparse

import requests
from rich.prompt import Confirm

import murfey.client.update
import murfey.client.watchdir
import murfey.client.websocket
from murfey.client.customlogging import CustomHandler, DirectableRichHandler
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.tui.app import MurfeyTUI
from murfey.client.tui.status_bar import StatusBar
from murfey.util.api import url_path_for
from murfey.util.client import authorised_requests, read_config
from murfey.util.models import Visit

log = logging.getLogger("murfey.client")

requests.get, requests.post, requests.put, requests.delete = authorised_requests()


def _get_visit_list(api_base: ParseResult, instrument_name: str):
    proxy_path = api_base.path.rstrip("/")
    get_visits_url = api_base._replace(
        path=f"{proxy_path}{url_path_for('session_control.router', 'get_current_visits', instrument_name=instrument_name)}"
    )
    server_reply = requests.get(get_visits_url.geturl())
    if server_reply.status_code != 200:
        raise ValueError(f"Server unreachable ({server_reply.status_code})")
    return [Visit.parse_obj(v) for v in server_reply.json()]


def write_config(config: configparser.ConfigParser):
    mcch = os.environ.get("MURFEY_CLIENT_CONFIG_HOME")
    murfey_client_config_home = Path(mcch) if mcch else Path.home()
    with open(murfey_client_config_home / ".murfey", "w") as configfile:
        config.write(configfile)


def main_loop(
    source_watchers: list[murfey.client.watchdir.DirWatcher],
    appearance_time: float,
    transfer_all: bool,
):
    log.info(
        f"Murfey {murfey.__version__} on Python {'.'.join(map(str, sys.version_info[0:3]))} entering main loop"
    )
    if appearance_time > 0:
        modification_time: float | None = time.time() - appearance_time * 3600
    else:
        modification_time = None
    while True:
        for sw in source_watchers:
            sw.scan(modification_time=modification_time, transfer_all=transfer_all)
        time.sleep(15)


def _enable_webbrowser_in_cygwin():
    """Helper function to make webbrowser.open() work in CygWin"""
    if "cygwin" in platform.system().lower() and shutil.which("cygstart"):
        webbrowser.register("cygstart", None, webbrowser.GenericBrowser("cygstart"))


def _check_for_updates(
    server: ParseResult, install_version: None | Literal[True] | str
):
    if install_version is True:
        # User requested installation of the newest version
        try:
            murfey.client.update.check(server, force=True)
            print("\nYou are already running the newest version of Murfey")
            exit()
        except Exception as e:
            exit(f"Murfey update check failed with {e}")

    if install_version:
        # User requested installation of a specific version
        if murfey.client.update.install_murfey(server, install_version):
            print(f"\nMurfey has been updated to version {install_version}")
            exit()
        else:
            exit("Error occurred while updating Murfey")

    # Otherwise run a routine update check to ensure client and server are compatible
    try:
        murfey.client.update.check(server)
    except Exception as e:
        print(f"Murfey update check failed with {e}")


def run():
    # Load client config and server information
    config = read_config()
    instrument_name = config["Murfey"]["instrument_name"]
    try:
        server_routing = config["ServerRouter"]
    except KeyError:
        server_routing = {}
    server_routing_prefix_found = False
    if server_routing:
        for path_prefix, server in server_routing.items():
            if str(Path.cwd()).startswith(path_prefix):
                known_server = server
                server_routing_prefix_found = True
                break
            else:
                known_server = None
    else:
        known_server = config["Murfey"].get("server")

    # Set up argument parser with dynamic defaults based on client config
    parser = argparse.ArgumentParser(description="Start the Murfey client")
    parser.add_argument(
        "--server",
        metavar="HOST:PORT",
        type=str,
        help=f"Murfey server to connect to ({known_server})",
        default=known_server,
    )
    parser.add_argument("--visit", help="Name of visit")
    parser.add_argument(
        "--source", help="Directory to transfer files from", type=Path, default="."
    )
    parser.add_argument(
        "--destination",
        help="Directory to transfer files to (syntax: 'data/2022/cm31093-2/tmp/murfey')",
    )
    parser.add_argument(
        "--update",
        metavar="VERSION",
        nargs="?",
        default=None,
        const=True,
        help="Update Murfey to the newest or to a specific version",
    )
    parser.add_argument(
        "--demo",
        action="store_true",
    )
    parser.add_argument(
        "--appearance-time",
        type=float,
        default=-1,
        help="Only consider top level directories that have appeared more recently than this many hours ago",
    )
    parser.add_argument(
        "--fake-dc",
        action="store_true",
        default=False,
        help="Do not perform data collection related calls to API (avoids database inserts)",
    )
    parser.add_argument(
        "--time-based-transfer",
        action="store_true",
        help="Transfer new files",
    )
    parser.add_argument(
        "--no-transfer",
        action="store_true",
        help="Avoid actually transferring files",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Turn on debugging logs",
    )
    parser.add_argument(
        "--local",
        action="store_true",
        default=False,
        help="Perform rsync transfers locally rather than remotely",
    )
    parser.add_argument(
        "--ignore-mdoc-metadata",
        action="store_true",
        default=False,
        help="Do not attempt to read metadata from all mdoc files",
    )
    parser.add_argument(
        "--remove-files",
        action="store_true",
        default=False,
        help="Remove source files immediately after their transfer",
    )
    parser.add_argument(
        "--name",
        type=str,
        default="",
        help="Name of Murfey session to be created",
    )
    parser.add_argument(
        "--skip-existing-processing",
        action="store_true",
        default=False,
        help="Do not trigger processing for any data directories currently on disk (you may have started processing for them in a previous murfey run)",
    )
    args = parser.parse_args()

    # Logic to exit early based on parsed args
    if not args.server:
        exit("Murfey server not set. Please run with --server host:port")
    if not args.server.startswith(("http://", "https://")):
        if "://" in args.server:
            exit("Unknown server protocol. Only http:// and https:// are allowed")
        args.server = f"http://{args.server}"
    if args.remove_files:
        remove_prompt = Confirm.ask(
            f"Are you sure you want to remove files from {args.source or Path('.').absolute()}?"
        )
        if not remove_prompt:
            exit("Exiting")

    # If a new server URL is provided, save info to config file
    murfey_url = urlparse(args.server, allow_fragments=False)
    if args.server != known_server:
        # New server specified. Verify that it is real
        print(f"Attempting to connect to new server {args.server}")
        try:
            murfey.client.update.check(murfey_url, install=False)
        except Exception as e:
            exit(f"Could not reach Murfey server at {args.server!r} - {e}")

        # If server is reachable then update the configuration
        config["Murfey"]["server"] = args.server
        write_config(config)

    # If user requested installation of a specific or a newer version then
    # make that happen, otherwise ensure client and server are compatible and
    # update if necessary.
    _check_for_updates(server=murfey_url, install_version=args.update)

    if args.no_transfer:
        log.info("No files will be transferred as --no-transfer flag was specified")

    # Check ISPyB (if set up) for ongoing visits
    ongoing_visits = []
    if args.visit:
        ongoing_visits = [args.visit]
    elif server_routing_prefix_found:
        for part in Path.cwd().parts:
            if "-" in part:
                ongoing_visits = [part]
                break
    if not ongoing_visits:
        print("Ongoing visits:")
        ongoing_visits = _get_visit_list(murfey_url, instrument_name)
        pprint(ongoing_visits)
        ongoing_visits = [v.name for v in ongoing_visits]

    _enable_webbrowser_in_cygwin()

    # Set up additional log handlers
    log.setLevel(logging.DEBUG)
    log_queue = Queue()
    input_queue = Queue()

    # Rich-based console handler
    rich_handler = DirectableRichHandler(enable_link_path=False)
    rich_handler.setLevel(logging.DEBUG if args.debug else logging.INFO)

    # Set up websocket app and handler
    client_id_response = requests.get(
        f"{murfey_url.geturl()}{url_path_for('session_control.router', 'new_client_id')}"
    )
    if client_id_response.status_code == 401:
        exit(
            "This instrument is not authorised to run the TUI app; please use the "
            "Murfey web UI instead"
        )
    elif client_id_response.status_code != 200:
        exit(
            "Unable to establish connection to Murfey server: \n"
            f"{client_id_response.json()}"
        )
    client_id: dict = client_id_response.json()
    ws = murfey.client.websocket.WSApp(
        server=args.server,
        id=client_id["new_id"],
    )
    ws_handler = CustomHandler(ws.send)

    # Add additional handlers and set logging levels
    logging.getLogger().addHandler(rich_handler)
    logging.getLogger().addHandler(ws_handler)
    logging.getLogger("murfey").setLevel(logging.INFO)
    logging.getLogger("websocket").setLevel(logging.WARNING)

    log.info("Starting Websocket connection")

    # Load machine data for subsequent sections
    machine_data = requests.get(
        f"{murfey_url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=instrument_name)}"
    ).json()
    gain_ref: Path | None = None

    # Set up Murfey environment instance and map it to websocket app
    instance_environment = MurfeyInstanceEnvironment(
        url=murfey_url,
        client_id=ws.id,
        instrument_name=instrument_name,
        software_versions=machine_data.get("software_versions", {}),
        demo=args.demo,
        processing_only_mode=server_routing_prefix_found,
        rsync_url=(
            urlparse(machine_data["rsync_url"]).hostname
            if machine_data.get("rsync_url")
            else ""
        ),
    )
    ws.environment = instance_environment

    # Set up and run Murfey TUI app
    status_bar = StatusBar()
    rich_handler.redirect = True
    app = MurfeyTUI(
        environment=instance_environment,
        visits=ongoing_visits,
        queues={"input": input_queue, "logs": log_queue},
        status_bar=status_bar,
        dummy_dc=args.fake_dc,
        do_transfer=not args.no_transfer,
        gain_ref=gain_ref,
        redirected_logger=rich_handler,
        force_mdoc_metadata=not args.ignore_mdoc_metadata,
        processing_enabled=machine_data.get("processing_enabled", True),
        skip_existing_processing=args.skip_existing_processing,
    )
    app.run()
    rich_handler.redirect = False



from __future__ import annotations

# import contextlib
import logging
import subprocess
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    OrderedDict,
    Type,
    TypeVar,
)

import requests
from pydantic import BaseModel, ValidationError
from rich.box import SQUARE
from rich.panel import Panel
from textual.app import ScreenStackError
from textual.containers import VerticalScroll
from textual.message import Message
from textual.reactive import reactive
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import (
    Button,
    DataTable,
    DirectoryTree,
    Footer,
    Header,
    Input,
    Label,
    ProgressBar,
    RadioButton,
    RadioSet,
    RichLog,
    Static,
    Switch,
    Tree,
)
from werkzeug.utils import secure_filename

from murfey.client.analyser import Analyser
from murfey.client.contexts.spa import SPAModularContext
from murfey.client.contexts.tomo import TomographyContext
from murfey.client.gain_ref import determine_gain_ref
from murfey.client.instance_environment import (
    MurfeyInstanceEnvironment,
    global_env_lock,
)
from murfey.client.rsync import RSyncer
from murfey.util import posix_path
from murfey.util.api import url_path_for
from murfey.util.client import capture_post, get_machine_config_client, read_config
from murfey.util.models import ProcessingParametersSPA, ProcessingParametersTomo

log = logging.getLogger("murfey.tui.screens")

ReactiveType = TypeVar("ReactiveType")

token = read_config()["Murfey"].get("token", "")
instrument_name = read_config()["Murfey"].get("instrument_name", "")

requests.get = partial(requests.get, headers={"Authorization": f"Bearer {token}"})
requests.post = partial(requests.post, headers={"Authorization": f"Bearer {token}"})
requests.delete = partial(requests.delete, headers={"Authorization": f"Bearer {token}"})


def determine_default_destination(
    visit: str,
    source: Path,
    destination: str,
    environment: MurfeyInstanceEnvironment,
    analysers: Dict[Path, Analyser],
    touch: bool = False,
    extra_directory: str = "",
    include_mid_path: bool = True,
    use_suggested_path: bool = True,
) -> str:
    machine_data = requests.get(
        f"{environment.url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=environment.instrument_name)}"
    ).json()
    _default = ""
    if environment.processing_only_mode and environment.sources:
        log.info(f"Processing only mode with sources {environment.sources}")
        _default = str(environment.sources[0].absolute()) or str(Path.cwd())
    elif machine_data.get("data_directories"):
        for data_dir in machine_data["data_directories"]:
            if source.absolute() == Path(data_dir).absolute():
                _default = f"{destination}/{visit}"
                break
            else:
                try:
                    mid_path = source.absolute().relative_to(Path(data_dir).absolute())
                    if use_suggested_path:
                        with global_env_lock:
                            source_name = (
                                source.name
                                if source.name != "Images-Disc1"
                                else source.parent.name
                            )
                            if environment.destination_registry.get(source_name):
                                _default = environment.destination_registry[source_name]
                            else:
                                suggested_path_response = capture_post(
                                    url=f"{str(environment.url.geturl())}{url_path_for('file_io_instrument.router', 'suggest_path', visit_name=visit, session_id=environment.murfey_session)}",
                                    json={
                                        "base_path": f"{destination}/{visit}/{mid_path.parent if include_mid_path else ''}/raw",
                                        "touch": touch,
                                        "extra_directory": extra_directory,
                                    },
                                )
                                if suggested_path_response is None:
                                    raise RuntimeError("Murfey server is unreachable")
                                _default = suggested_path_response.json().get(
                                    "suggested_path"
                                )
                                environment.destination_registry[source_name] = _default
                    else:
                        _default = f"{destination}/{visit}/{mid_path if include_mid_path else source.name}"
                    break
                except (ValueError, KeyError):
                    _default = ""
        else:
            _default = ""
    else:
        _default = f"{destination}/{visit}"
    return (
        _default + f"/{extra_directory}"
        if not _default.endswith("/")
        else _default + f"{extra_directory}"
    )


class InputResponse(NamedTuple):
    question: str
    allowed_responses: List[str] | None = None
    default: str = ""
    callback: Callable | None = None
    key_change_callback: Callable | None = None
    kwargs: dict | None = None
    form: OrderedDict[str, Any] | None = None
    model: BaseModel | None = None


class LogBook(RichLog):
    class Log(Message):
        def __init__(self, log_renderable):
            self.renderable = log_renderable
            super().__init__()


class InfoWidget(Widget):
    text: reactive[str] = reactive("")

    def __init__(self, text: str, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def render(self) -> Panel:
        return Panel(self.text, style=("on dark_magenta"), box=SQUARE)

    def _key_change(self, input_char: str | None):
        if input_char is None:
            self.text = self.text[:-1]
            return
        self.text += input_char


class QuickPrompt:
    def __init__(self, text: str, options: List[str]):
        self._text = text
        self._options = options
        self.warn = False

    def __repr__(self):
        return repr(self._text)

    def __str__(self):
        return self._text

    def __iter__(self):
        return iter(self._options)

    def __bool__(self):
        return bool(self._text)


def validate_form(form: dict, model: BaseModel) -> bool:
    try:
        convert = lambda x: None if x == "None" else x
        validated = model(**{k: convert(v) for k, v in form.items()})
        log.info(validated.model_dump())
        return True
    except (AttributeError, ValidationError) as e:
        log.warning(f"Form validation failed: {str(e)}")
        return False


class _DirectoryTree(DirectoryTree):
    valid_selection = reactive(False)

    def __init__(self, *args, data_directories: List[Path] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._selected_path = self.path
        self._data_directories = data_directories or []

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        event.stop()
        dir_entry = event.node.data.path
        if dir_entry is None:
            return
        if dir_entry.is_dir():
            self._selected_path = dir_entry
            if not self._data_directories:
                self.valid_selection = True
                return
            for d in self._data_directories:
                if Path(self._selected_path).absolute().is_relative_to(d.absolute()):
                    self.valid_selection = True
                    break
            else:
                self.valid_selection = False
        else:
            self.valid_selection = False


class _DirectoryTreeGain(DirectoryTree):
    valid_selection = reactive(False)

    def __init__(self, gain_reference: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._gain_reference = gain_reference

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        event.stop()
        dir_entry = event.node.data.path
        if dir_entry is None:
            return
        if not dir_entry.is_dir():
            self.valid_selection = True
            self._gain_reference = dir_entry
        else:
            self.valid_selection = False


class LaunchScreen(Screen):
    _launch_btn: Button | None = None

    def __init__(
        self, *args, basepath: Path = Path("."), add_basepath: bool = False, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._selected_dir = basepath
        self._add_basepath = add_basepath
        self._context: Type[SPAModularContext] | Type[TomographyContext]
        self._context = SPAModularContext

    def compose(self):

        machine_data = requests.get(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=instrument_name)}"
        ).json()
        self._dir_tree = _DirectoryTree(
            str(self._selected_dir),
            data_directories=machine_data.get("data_directories", []),
            id="dir-select",
        )

        yield self._dir_tree
        text_log = RichLog(id="selected-directories")
        widgets = [text_log, Button("Clear", id="clear")]
        text_log_block = VerticalScroll(*widgets, id="selected-directories-vert")
        yield text_log_block

        text_log.write("Selected directories:\n")
        btn_disabled = True
        for d in machine_data.get("data_directories", []):
            if (
                Path(self._dir_tree._selected_path)
                .absolute()
                .is_relative_to(Path(d).absolute())
                or self.app._environment.processing_only_mode
            ):
                btn_disabled = False
                break
        self._launch_btn = Button("Launch", id="launch", disabled=btn_disabled)
        self._add_btn = Button("Add directory", id="add", disabled=btn_disabled)
        self.watch(self._dir_tree, "valid_selection", self._check_valid_selection)
        yield self._add_btn
        yield self._launch_btn
        yield Button("Quit", id="quit")

    def on_mount(self):
        if self._add_basepath:
            self._add_directory(str(self._selected_dir))

    def _check_valid_selection(self, valid: bool):
        if self._add_btn:
            if valid:
                self._add_btn.disabled = False
            else:
                self._add_btn.disabled = True

    def _add_directory(self, directory: str, add_destination: bool = True):
        source = Path(self._dir_tree.path).absolute() / directory
        if add_destination:
            for s in self.app._environment.sources:
                if source.is_relative_to(s):
                    return
            self.app._environment.sources.append(source)
            self.app._default_destinations[source] = f"{datetime.now().year}"
        if self._launch_btn:
            self._launch_btn.disabled = False
        self.query_one("#selected-directories").write(str(source) + "\n")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "quit":
            self.app.clean_up_quit()
        elif event.button.id == "add":
            self._add_directory(self._dir_tree._selected_path)
        elif event.button.id == "launch":
            text = self.app._visit
            visit_path = ""
            transfer_routes = {}
            for s, defd in self.app._default_destinations.items():
                _default = determine_default_destination(
                    self.app._visit,
                    s,
                    defd,
                    self.app._environment,
                    self.app.analysers,
                    touch=True,
                )
                visit_path = defd + f"/{text}"
                if self.app._environment.processing_only_mode:
                    self.app._start_rsyncer(
                        Path(_default), _default, visit_path=visit_path
                    )
                transfer_routes[s] = _default
            self.app.install_screen(
                DestinationSelect(transfer_routes, self._context),
                "destination-select-screen",
            )
            self.app.pop_screen()
            self.app.push_screen("destination-select-screen")
        elif event.button.id == "clear":
            sel_dir = self.query_one("#selected-directories")
            for line in sel_dir.lines[1:]:
                source = Path(line.text)
                if source in self.app._environment.sources:
                    self.app._environment.sources.remove(source)
                    if self.app._default_destinations.get(source):
                        del self.app._default_destinations[source]
            sel_dir.clear()
            sel_dir.write("Selected directories:\n")


class ConfirmScreen(Screen):
    def __init__(
        self,
        prompt: str,
        *args,
        params: dict | None = None,
        pressed_callback: Callable | None = None,
        button_names: dict | None = None,
        push: str = "main",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._prompt = prompt
        self._params = params or {}
        self._callback = pressed_callback
        self._button_names = button_names or {}
        self._push = push

    def compose(self):
        if self._params:
            dt = DataTable(id="prompt")
            keys = list(self._params.keys())
            dt.add_columns(*keys)
            dt.add_rows([[self._params[k] for k in keys]])
            yield dt
        else:
            yield Static(self._prompt, id="prompt")
        yield Button(self._button_names.get("launch") or "Launch", id="launch")
        yield Button(self._button_names.get("quit") or "Back", id="quit")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "quit":
            self.app.pop_screen()
            self.app.uninstall_screen("confirm")
        else:
            while True:
                try:
                    if self.app.screen._name == "main":
                        break
                    self.app.pop_screen()
                except ScreenStackError:
                    break
            if self._push:
                log.info(f"Pushing screen {self._push}")
                self.app.push_screen(self._push)
            self.app.uninstall_screen("confirm")
        if self._callback and event.button.id == "launch":
            self._callback(params=self._params)


class ProcessingForm(Screen):
    _form = reactive({})
    _vert = None

    def __init__(
        self,
        form: dict,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._form = form
        self._inputs: Dict[Input, str] = {}

    def compose(self):
        inputs = []
        analyser = list(self.app.analysers.values())[0]
        for k in analyser._context.user_params + analyser._context.metadata_params:
            t = k.label
            inputs.append(Label(t, classes="label"))
            if self._form.get(k.name) in ("true", "True", True):
                i = Switch(value=True, classes="input", id=f"switch_{k.name}")
            elif self._form.get(k.name) in ("false", "False", False):
                i = Switch(value=False, classes="input", id=f"switch_{k.name}")
            else:
                i = Input(placeholder=t, classes="input", id=f"input_{k.name}")
                default = self._form.get(k.name, str(k.default))
                i.value = "None" if default is None else default
            self._inputs[i] = k.name
            inputs.append(i)
        confirm_btn = Button("Confirm", id="confirm-btn")
        if self._form.get("motion_corr_binning") == "2":
            self._vert = VerticalScroll(
                *inputs,
                Label("Collected in counting mode:"),
                Switch(id="superres", value=True, classes="input"),
                confirm_btn,
                id="input-form",
            )
        else:
            self._vert = VerticalScroll(*inputs, confirm_btn, id="input-form")
        yield self._vert

    def _write_params(
        self,
        params: dict | None = None,
        model: ProcessingParametersTomo | ProcessingParametersSPA | None = None,
    ):
        if params:
            try:
                analyser = [a for a in self.app.analysers.values() if a._context][0]
            except IndexError:
                return
            for k in analyser._context.user_params + analyser._context.metadata_params:
                self.app.query_one("#info").write(f"{k.label}: {params.get(k.name)}")
            self.app._start_dc(params)
            if model == ProcessingParametersTomo:
                requests.post(
                    f"{self.app._environment.url.geturl()}/{url_path_for('workflow.tomo_router', 'register_tomo_proc_params', session_id=self.app._environment.murfey_session)}",
                    json=params,
                )
            elif model == ProcessingParametersSPA:
                requests.post(
                    f"{self.app._environment.url.geturl()}{url_path_for('workflow.spa_router', 'register_spa_proc_params', session_id=self.app._environment.murfey_session)}",
                    json=params,
                )
                requests.post(
                    f"{self.app._environment.url.geturl()}{url_path_for('workflow.spa_router', 'flush_spa_processing', visit_name=self.app._environment.visit, session_id=self.app._environment.murfey_session)}",
                )

    def on_switch_changed(self, event):
        if event.switch.id == "superres":
            pix_size = self.query_one("#input_pixel_size_on_image")
            motion_corr_binning = self.query_one("#input_motion_corr_binning")
            if event.value:
                pix_size.value = str(float(pix_size.value) / 2)
                motion_corr_binning.value = "2"
            else:
                pix_size.value = str(float(pix_size.value) * 2)
                motion_corr_binning.value = "1"
        else:
            k = self._inputs[event.switch]
            self._form[k] = event.value

    def on_input_changed(self, event):
        k = self._inputs[event.input]
        self._form[k] = event.value

    def on_button_pressed(self, event):
        model = None
        if self.app.analysers.get(Path(self._form.get("source", ""))):
            if model := self.app.analysers[Path(self._form["source"])].parameters_model:
                valid = validate_form(self._form, model)
                if not valid:
                    return
        if "confirm" not in self.app._installed_screens:
            self.app.install_screen(
                ConfirmScreen(
                    "Launch processing?",
                    params=self._form,
                    pressed_callback=partial(self._write_params, model=model),
                ),
                "confirm",
            )
        self.app.push_screen("confirm")


class SwitchSelection(Screen):
    def __init__(
        self,
        name: str,
        elements: List[str],
        switch_label: str,
        switch_status: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._elements = elements
        self._switch_status = switch_status
        self._switch_label = switch_label
        self._name = name

    def compose(self):
        hovers = (
            [
                Button(e, id=f"btn-{self._name}-{e}", classes=f"btn-{self._name}")
                for e in self._elements
            ]
            if self._elements
            else [Button("No elements found")]
        )
        yield VerticalScroll(*hovers, id=f"select-{self._name}")
        yield Static(self._switch_label, id=f"label-{self._name}")
        yield Switch(id=f"switch-{self._name}", value=self._switch_status)

    def on_switch_changed(self, event):
        self._switch_status = event.value


class SessionSelection(Screen):
    def __init__(
        self, sessions: List[str], sessions_with_client: List[str], *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._sessions = sessions
        self._sessions_with_client = sessions_with_client
        self._name = "session"

    def compose(self):
        hovers = (
            [
                Button(e, id=f"btn-{self._name}-{e}", classes=f"btn-{self._name}")
                for e in self._sessions
            ]
            if self._sessions
            else [Button("No elements found")]
        )
        deletes = (
            [
                Button(
                    f"Remove {e}",
                    id=f"btn-{self._name}-{e}-del",
                    classes=f"btn-{self._name}",
                )
                for e in self._sessions
            ]
            if self._sessions
            else [Button("No elements found")]
        )
        yield VerticalScroll(
            *[v for pair in zip(hovers, deletes) for v in pair],
            id=f"select-{self._name}",
        )
        yield Button(
            "New session", id=f"btn-{self._name}-new", classes=f"btn-{self._name}"
        )

    def on_button_pressed(self, event: Button.Pressed):
        if event.button.id.endswith("new"):
            session_id = None
            self.app.pop_screen()
        elif event.button.id.endswith("del"):
            session_id = int(
                str(event.button.label.split(":")[0]).replace("Remove ", "")
            )
            self.app.pop_screen()
            self.app.install_screen(
                ConfirmScreen(
                    (
                        f"Remove session {session_id} [WARNING: there are clients already using this session]"
                        if str(event.button.label) in self._sessions_with_client
                        else f"Remove session {session_id}"
                    ),
                    pressed_callback=partial(self._remove_session, session_id),
                    button_names={"launch": "Yes"},
                    push="visit-select-screen",
                ),
                "confirm",
            )
            self.app.push_screen("confirm")
            return
        else:
            self.app._environment.murfey_session = int(
                str(event.button.label.split(":")[0])
            )
            session_id = self.app._environment.murfey_session
            self.app.pop_screen()
        session_name = "Client connection"
        self.app._environment.murfey_session = requests.post(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'link_client_to_session', instrument_name=self.app._environment.instrument_name, client_id=self.app._environment.client_id)}",
            json={"session_id": session_id, "session_name": session_name},
        ).json()

    def _remove_session(self, session_id: int, **kwargs):
        requests.delete(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'remove_session', session_id=session_id)}"
        )
        exisiting_sessions = requests.get(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'get_sessions')}"
        ).json()
        self.app.uninstall_screen("session-select-screen")
        if exisiting_sessions:
            self.app.install_screen(
                SessionSelection(
                    [
                        f"{s['session']['id']}: {s['session']['name']}"
                        for s in exisiting_sessions
                    ],
                    [
                        f"{s['session']['id']}: {s['session']['name']}"
                        for s in exisiting_sessions
                        if s["clients"]
                    ],
                ),
                "session-select-screen",
            )
            self.app.push_screen("session-select-screen")
        else:
            session_name = "Client connection"
            resp = capture_post(
                f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'link_client_to_session', instrument_name=self.app._environment.instrument_name, client_id=self.app._environment.client_id)}",
                json={"session_id": None, "session_name": session_name},
            )
            if resp:
                self.app._environment.murfey_session = resp.json()


class VisitSelection(SwitchSelection):
    def __init__(self, visits: List[str], *args, **kwargs):
        super().__init__(
            "visit",
            visits,
            "Create visit directory (suggested)",
            *args,
            **kwargs,
        )

    def on_button_pressed(self, event: Button.Pressed):
        text = str(event.button.label)
        self.app._visit = text
        self.app._environment.visit = text
        response = requests.post(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'register_client_to_visit', visit_name=text)}",
            json={"id": self.app._environment.client_id},
        )
        log.info(f"Posted visit registration: {response.status_code}")
        machine_data = requests.get(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=instrument_name)}"
        ).json()

        if self._switch_status:
            self.app.install_screen(
                DirectorySelection(
                    [
                        path
                        for path in machine_data.get("data_directories", [])
                        if Path(path).exists()
                    ]
                ),
                "directory-select",
            )
        self.app.pop_screen()

        if machine_data.get("gain_reference_directory"):
            self.app.install_screen(
                GainReference(
                    determine_gain_ref(Path(machine_data["gain_reference_directory"])),
                    self._switch_status,
                ),
                "gain-ref-select",
            )
            self.app.push_screen("gain-ref-select")
        else:
            if self._switch_status:
                self.app.push_screen("directory-select")
            else:
                self.app.install_screen(LaunchScreen(basepath=Path("./")), "launcher")
                self.app.push_screen("launcher")

        if machine_data.get("upstream_data_directories"):
            upstream_downloads = requests.get(
                f"{self.app._environment.url.geturl()}{url_path_for('session_control.correlative_router', 'find_upstream_visits', session_id=self.app._environment.murfey_session)}"
            ).json()
            self.app.install_screen(
                UpstreamDownloads(upstream_downloads), "upstream-downloads"
            )
            self.app.push_screen("upstream-downloads")


class VisitCreation(Screen):
    # This allows for the manual creation of a visit name when there is no LIMS system to provide it
    # Shares a lot of code with VisitSelection, should be neatened up at some point
    visit_name: reactive[str] = reactive("")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compose(self):
        yield Input(placeholder="Visit name", classes="input-visit-name")
        yield Button("Create visit", classes="btn-visit-create")

    def on_input_changed(self, event):
        self.visit_name = event.value

    def on_button_pressed(self, event: Button.Pressed):
        text = str(self.visit_name)
        self.app._visit = text
        self.app._environment.visit = text
        response = requests.post(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'register_client_to_visit', visit_name=text)}",
            json={"id": self.app._environment.client_id},
        )
        log.info(f"Posted visit registration: {response.status_code}")
        machine_data = requests.get(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=instrument_name)}"
        ).json()

        self.app.install_screen(
            DirectorySelection(
                [
                    path
                    for path in machine_data.get("data_directories", [])
                    if Path(path).exists()
                ]
            ),
            "directory-select",
        )
        self.app.pop_screen()

        if machine_data.get("gain_reference_directory"):
            self.app.install_screen(
                GainReference(
                    determine_gain_ref(Path(machine_data["gain_reference_directory"])),
                    True,
                ),
                "gain-ref-select",
            )
            self.app.push_screen("gain-ref-select")
        else:
            self.app.push_screen("directory-select")

        if machine_data.get("upstream_data_directories"):
            upstream_downloads = requests.get(
                f"{self.app._environment.url.geturl()}{url_path_for('session_control.correlative_router', 'find_upstream_visits', session_id=self.app._environment.murfey_session)}"
            ).json()
            self.app.install_screen(
                UpstreamDownloads(upstream_downloads), "upstream-downloads"
            )
            self.app.push_screen("upstream-downloads")


class UpstreamDownloads(Screen):
    def __init__(self, connected_visits: Dict[str, Path], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connected_visits = connected_visits

    def compose(self):
        visit_buttons = [
            Button(cv, classes="btn-directory") for cv in self._connected_visits.keys()
        ]
        yield VerticalScroll(*visit_buttons)
        yield Button("Skip", classes="btn-directory")

    def on_button_pressed(self, event: Button.Pressed):
        machine_data = requests.get(
            f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=instrument_name)}"
        ).json()
        if machine_data.get("upstream_data_download_directory"):
            # Create the directory locally to save files to
            download_dir = Path(machine_data["upstream_data_download_directory"]) / str(
                event.button.label
            )
            download_dir.mkdir(exist_ok=True)

            # Get the paths to the TIFF files generated previously under the same session ID
            upstream_tiff_paths_response = requests.get(
                f"{self.app._environment.url.geturl()}{url_path_for('session_control.correlative_router', 'gather_upstream_tiffs', visit_name=event.button.label, session_id=self.app._environment.murfey_session)}"
            )
            upstream_tiff_paths = upstream_tiff_paths_response.json() or []

            # Request to download the TIFF files found
            for tp in upstream_tiff_paths:
                (download_dir / tp).parent.mkdir(exist_ok=True, parents=True)
                # Write TIFF to the specified file path
                stream_response = requests.get(
                    f"{self.app._environment.url.geturl()}{url_path_for('session_control.correlative_router', 'get_tiff', visit_name=event.button.label, session_id=self.app._environment.murfey_session, tiff_path=tp)}",
                    stream=True,
                )
                # Write the file chunk-by-chunk to avoid hogging memory
                with open(download_dir / tp, "wb") as utiff:
                    for chunk in stream_response.iter_content(chunk_size=32 * 1024**2):
                        utiff.write(chunk)
        self.app.pop_screen()


class GainReference(Screen):
    def __init__(self, gain_reference: Path, switch_status: bool, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._gain_reference = gain_reference
        self._switch_status = switch_status

    def compose(self):
        self._dir_tree = _DirectoryTreeGain(
            self._gain_reference,
            str(self._gain_reference.parent.parent),
            id="gain-select",
        )
        yield self._dir_tree
        self._launch_btn = Button("Launch", id="launch")
        self.watch(self._dir_tree, "valid_selection", self._check_valid_selection)
        yield Button(
            f"Suggested gain reference: {self._gain_reference.parent / self._gain_reference.name}",
            id="suggested-gain-ref",
        )
        yield self._launch_btn
        yield Button("No gain", id="skip-gain")

    def _check_valid_selection(self, valid: bool):
        if self._launch_btn:
            if valid:
                self._launch_btn.disabled = False
            else:
                self._launch_btn.disabled = True

    def on_button_pressed(self, event):
        if event.button.id == "skip-gain":
            self.app.pop_screen()
        else:
            if event.button.id == "suggested-gain-ref":
                self._dir_tree._gain_reference = self._gain_reference
            visit_path = f"{datetime.now().year}/{self.app._environment.visit}"
            # Set up rsync command
            rsync_cmd = [
                "rsync",
                f"{posix_path(self._dir_tree._gain_reference)!r}",
                f"{self.app._environment.rsync_url or self.app._environment.url.hostname}::{self.app._machine_config.get('rsync_module', 'data')}/{visit_path}/processing/{secure_filename(self._dir_tree._gain_reference.name)}",
            ]
            # Encase in bash shell
            cmd = ["bash", "-c", " ".join(rsync_cmd)]
            if self.app._environment.demo:
                log.info(f"Would perform {' '.join(cmd)}")
            else:
                # Run rsync subprocess
                gain_rsync = subprocess.run(cmd)
                if gain_rsync.returncode:
                    log.warning(
                        f"Gain reference file {posix_path(self._dir_tree._gain_reference)!r} was not successfully transferred to {visit_path}/processing"
                    )
            process_gain_response = requests.post(
                url=f"{str(self.app._environment.url.geturl())}{url_path_for('file_io_instrument.router', 'process_gain', session_id=self.app._environment.murfey_session)}",
                json={
                    "gain_ref": str(self._dir_tree._gain_reference),
                    "eer": bool(
                        self.app._machine_config.get("external_executables_eer")
                    ),
                },
            )
            if str(process_gain_response.status_code).startswith("4"):
                log.warning(
                    f"Gain processing failed: status code {process_gain_response.status_code}"
                )
            else:
                log.info(
                    f"Gain reference file {process_gain_response.json().get('gain_ref')}"
                )
                self.app._environment.gain_ref = process_gain_response.json().get(
                    "gain_ref"
                )
        if self._switch_status:
            self.app.push_screen("directory-select")
        else:
            self.app.install_screen(LaunchScreen(basepath=Path("./")), "launcher")
            self.app.push_screen("launcher")


class DirectorySelection(SwitchSelection):
    def __init__(self, directories: List[str], *args, **kwargs):
        super().__init__(
            "directory",
            directories,
            "Automatically transfer and trigger processing for new directories (recommended)",
            *args,
            **kwargs,
        )

    def on_button_pressed(self, event: Button.Pressed):
        self.app._multigrid = self._switch_status
        visit_dir = Path(str(event.button.label)).absolute() / self.app._visit
        visit_dir.mkdir(exist_ok=True)
        self.app._set_default_acquisition_directories(visit_dir)
        machine_config = get_machine_config_client(
            str(self.app._environment.url.geturl()),
            instrument_name=self.app._environment.instrument_name,
            demo=self.app._environment.demo,
        )
        for dir in machine_config["create_directories"]:
            (visit_dir / dir).mkdir(exist_ok=True)
        self.app.install_screen(
            LaunchScreen(basepath=visit_dir, add_basepath=True), "launcher"
        )
        self.app.pop_screen()
        self.app.push_screen("launcher")


class DestinationSelect(Screen):
    def __init__(
        self,
        transfer_routes: Dict[Path, str],
        context: Type[SPAModularContext] | Type[TomographyContext],
        *args,
        destination_overrides: Optional[Dict[Path, str]] = None,
        use_transfer_routes: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._transfer_routes = transfer_routes
        self._destination_overrides: Dict[Path, str] = destination_overrides or {}
        self._user_params: Dict[str, str] = {}
        self._inputs: Dict[Input, str] = {}
        self._context = context
        self._use_transfer_routes = use_transfer_routes

    def compose(self):
        bulk = []
        with RadioSet():
            yield RadioButton("SPA", value=self._context is SPAModularContext)
            yield RadioButton("Tomography", value=self._context is TomographyContext)
        if self.app._multigrid:
            machine_config = get_machine_config_client(
                str(self.app._environment.url.geturl()),
                instrument_name=self.app._environment.instrument_name,
            )
            destinations = []
            if self._destination_overrides:
                for k, v in self._destination_overrides.items():
                    destinations.append(v)
                    bulk.append(Label(f"Copy the source {k} to:"))
                    bulk.append(
                        Input(
                            value=v,
                            id=f"destination-{str(k)}",
                            classes="input-destination",
                        )
                    )
            else:
                for s in self._transfer_routes.keys():
                    for d in s.glob("*"):
                        if (
                            d.is_dir()
                            and d.name not in machine_config["create_directories"]
                        ):
                            dest = determine_default_destination(
                                self.app._visit,
                                s,
                                f"{datetime.now().year}",
                                self.app._environment,
                                self.app.analysers,
                                touch=True,
                            )
                            if dest and dest in destinations:
                                dest_path = Path(dest)
                                name_root = ""
                                dest_num = 0
                                for i, st in enumerate(dest_path.name):
                                    if st.isnumeric():
                                        dest_num = int(dest_path.name[i:])
                                        break
                                    name_root += st
                                if dest_num:
                                    dest = str(
                                        dest_path.parent / f"{name_root}{dest_num+1}"
                                    )
                                else:
                                    dest = str(dest_path.parent / f"{name_root}2")
                            destinations.append(dest)
                            bulk.append(Label(f"Copy the source {d} to:"))
                            bulk.append(
                                Input(
                                    value=dest,
                                    id=f"destination-{str(d)}",
                                    classes="input-destination",
                                )
                            )
        else:
            machine_config = get_machine_config_client(
                str(self.app._environment.url.geturl()),
                instrument_name=self.app._environment.instrument_name,
            )
            for s, d in self._transfer_routes.items():
                if Path(d).name not in machine_config["create_directories"]:
                    bulk.append(Label(f"Copy the source {s} to:"))
                    bulk.append(
                        Input(
                            value=d,
                            id=f"destination-{str(s)}",
                            classes="input-destination",
                        )
                    )
        yield VerticalScroll(*bulk, id="destination-holder")
        params_bulk = []
        if self.app._multigrid and self.app._processing_enabled:
            for k in self._context.user_params:
                params_bulk.append(Label(k.label))
                val = getattr(self.app._environment, k.name, str(k.default))
                self._user_params[k.name] = val
                if val in ("true", "True", True):
                    i = Switch(value=True, id=k.name, classes="input-destination")
                elif val in ("false", "False", False):
                    i = Switch(value=False, id=k.name, classes="input-destination")
                else:
                    i = Input(value=val, id=k.name, classes="input-destination")
                params_bulk.append(i)
                self._inputs[i] = k.name
            machine_config = get_machine_config_client(
                str(self.app._environment.url.geturl()),
                instrument_name=self.app._environment.instrument_name,
                demo=self.app._environment.demo,
            )
            if machine_config.get("superres"):
                params_bulk.append(
                    Label("Collected in super resoultion mode unbinned:")
                )
                params_bulk.append(
                    Switch(
                        value=False,
                        id="superres-multigrid",
                        classes="input-destination",
                    )
                )
                self.app._environment.superres = False
        yield VerticalScroll(
            *params_bulk,
            id="user-params",
        )
        yield Button("Confirm", id="destination-btn")

    def on_switch_changed(self, event):
        if event.switch.id == "superres-multigrid":
            self.app._environment.superres = event.value
        else:
            for k in self._context.user_params:
                if event.switch.id == k.name:
                    self._user_params[k.name] = event.value

    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        if event.index == 0:
            self._context = SPAModularContext
        else:
            self._context = TomographyContext
        self.app.pop_screen()
        self.app.uninstall_screen("destination-select-screen")
        self.app.install_screen(
            DestinationSelect(
                self._transfer_routes,
                self._context,
                destination_overrides=self._destination_overrides,
                use_transfer_routes=True,
            ),
            "destination-select-screen",
        )
        self.app.push_screen("destination-select-screen")

    def on_input_changed(self, event):
        if event.input.id.startswith("destination-"):
            if not self.app._multigrid:
                self._transfer_routes[Path(event.input.id[12:])] = event.value
            else:
                self._destination_overrides[Path(event.input.id[12:])] = event.value
        else:
            for k in self._context.user_params:
                if event.input.id == k.name:
                    self._user_params[k.name] = event.value

    def on_button_pressed(self, event):
        if self.app._multigrid and self.app._processing_enabled:
            if self._context == TomographyContext:
                valid = validate_form(self._user_params, ProcessingParametersTomo.Base)
            else:
                valid = validate_form(self._user_params, ProcessingParametersSPA.Base)
            if not valid:
                return
        for s, d in self._transfer_routes.items():
            self.app._default_destinations[s] = d
            self.app._register_dc = True
            if self.app._multigrid:
                for k, v in self._destination_overrides.items():
                    self.app._environment.destination_registry[k.name] = v
                self.app._launch_multigrid_watcher(
                    s, destination_overrides=self._destination_overrides
                )
            else:
                self.app._start_rsyncer(s, d)
        for k, v in self._user_params.items():
            setattr(self.app._environment, k, v)
        self.app.pop_screen()
        self.app.push_screen("main")


class WaitingScreen(Screen):
    def __init__(
        self,
        prompt: str,
        num_files: int,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._num_files = 0
        self._prompt = prompt
        self._new_rsyncers: List[RSyncer] = []

    def compose(self):
        yield Static(self._prompt, id="waiting-prompt")
        yield ProgressBar(id="progress")
        yield Button("Start", id="waiting-launch")
        yield Button("Back", id="waiting-quit")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "waiting-quit":
            self.app.pop_screen()
            self.app.uninstall_screen("waiting")
        else:
            sources = []
            if self.app.rsync_processes or self.app._environment.demo:
                for k, rp in self.app.rsync_processes.items():
                    rp.stop()
                    if self.app.analysers.get(k):
                        self.app.analysers[k].stop()
                    removal_rp = RSyncer.from_rsyncer(rp, remove_files=True)
                    removal_rp._listeners = [self.file_copied]
                    self._new_rsyncers.append(removal_rp)
                    sources.append(k)
            log.info(
                f"Starting to remove data files {self.app._environment.demo}, {len(self.app.rsync_processes)}"
            )
            self.query_one(ProgressBar).update(total=self._num_files)
            for k, removal_rp in zip(sources, self._new_rsyncers):
                for f in k.absolute().glob("**/*"):
                    if f.is_file():
                        self._num_files += 1
                        removal_rp.queue.put(f)
                removal_rp.queue.put(None)
            self.query_one(ProgressBar).update(total=self._num_files)
            for removal_rp in self._new_rsyncers:
                removal_rp.start()

    def file_copied(self, *args, **kwargs):
        self.query_one(ProgressBar).advance(1)
        if self.query_one(ProgressBar).progress == self.query_one(ProgressBar).total:
            requests.post(
                f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'register_processing_success_in_ispyb', session_id=self.app._environment.murfey_session)}"
            )
            requests.delete(
                f"{self.app._environment.url.geturl()}{url_path_for('session_control.router', 'remove_session', session_id=self.app._environment.murfey_session)}"
            )
            self.app.exit()


class MainScreen(Screen):
    def compose(self):
        self.app.log_book = LogBook(id="log_book", wrap=True, max_lines=200)
        if self.app._redirected_logger:
            log.info("connecting logger")
            self.app._redirected_logger.text_log = self.app.log_book
            log.info("logger connected")
        self.app.hovers = (
            [Button(v, id="visit-btn") for v in self.app.visits]
            if len(self.app.visits)
            else [Button("No ongoing visits found")]
        )
        self.app.processing_form = ProcessingForm(self.app._form_values)
        yield Header()
        info_widget = RichLog(id="info", markup=True)
        yield info_widget
        yield self.app.log_book
        info_widget.write(
            f"[bold]Welcome to Murfey ({self.app._environment.visit})[/bold]"
        )
        yield Button("Visit complete", id="new-visit-btn")
        yield Footer()

    def on_mount(self, event):
        requests.post(
            f"{self.app._environment.url.geturl()}{url_path_for('prometheus.router', 'change_monitoring_status', visit_name=self.app._environment.visit, on=1)}"
        )



import logging
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, Optional

from sqlmodel import select
from werkzeug.utils import secure_filename

import murfey.server.prometheus as prom
from murfey.util import safe_run, sanitise
from murfey.util.config import MachineConfig, from_file, get_machine_config, settings
from murfey.util.db import (
    DataCollection,
    DataCollectionGroup,
    FoilHole,
    GridSquare,
    ProcessingJob,
    RsyncInstance,
    Session,
)

logger = logging.getLogger("murfey.server.api.shared")


@lru_cache(maxsize=5)
def get_machine_config_for_instrument(instrument_name: str) -> Optional[MachineConfig]:
    if settings.murfey_machine_configuration:
        return from_file(Path(settings.murfey_machine_configuration), instrument_name)[
            instrument_name
        ]
    return None


def remove_session_by_id(session_id: int, db):
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    sessions_for_visit = db.exec(
        select(Session).where(Session.visit == session.visit)
    ).all()
    # Don't remove prometheus metrics if there are other sessions using them
    if len(sessions_for_visit) == 1:
        safe_run(
            prom.monitoring_switch.remove,
            args=(session.visit,),
            label="monitoring_switch",
        )
        rsync_instances = db.exec(
            select(RsyncInstance).where(RsyncInstance.session_id == session_id)
        ).all()
        for ri in rsync_instances:
            safe_run(
                prom.seen_files.remove,
                args=(ri.source, session.visit),
                label="seen_files",
            )
            safe_run(
                prom.transferred_files.remove,
                args=(ri.source, session.visit),
                label="transferred_files",
            )
            safe_run(
                prom.transferred_files_bytes.remove,
                args=(ri.source, session.visit),
                label="transferred_files_bytes",
            )
            safe_run(
                prom.seen_data_files.remove,
                args=(ri.source, session.visit),
                label="seen_data_files",
            )
            safe_run(
                prom.transferred_data_files.remove,
                args=(ri.source, session.visit),
                label="transferred_data_files",
            )
            safe_run(
                prom.transferred_data_files_bytes.remove,
                args=(ri.source, session.visit),
                label="transferred_data_file_bytes",
            )
    collected_ids = db.exec(
        select(DataCollectionGroup, DataCollection, ProcessingJob)
        .where(DataCollectionGroup.session_id == session_id)
        .where(DataCollection.dcg_id == DataCollectionGroup.id)
        .where(ProcessingJob.dc_id == DataCollection.id)
    ).all()
    for c in collected_ids:
        safe_run(
            prom.preprocessed_movies.remove,
            args=(c[2].id,),
            label="preprocessed_movies",
        )
    db.delete(session)
    db.commit()
    logger.debug(f"Successfully removed session {session_id} from database")
    return


def get_grid_squares(session_id: int, db):
    grid_squares = db.exec(
        select(GridSquare).where(GridSquare.session_id == session_id)
    ).all()
    tags = {gs.tag for gs in grid_squares}
    res = {}
    for t in tags:
        res[t] = [gs for gs in grid_squares if gs.tag == t]
    return res


def get_grid_squares_from_dcg(session_id: int, dcgid: int, db) -> List[GridSquare]:
    grid_squares = db.exec(
        select(GridSquare, DataCollectionGroup)
        .where(GridSquare.session_id == session_id)
        .where(GridSquare.tag == DataCollectionGroup.tag)
        .where(DataCollectionGroup.id == dcgid)
    ).all()
    return [gs[0] for gs in grid_squares]


def get_foil_holes_from_grid_square(
    session_id: int, dcgid: int, gsid: int, db
) -> List[FoilHole]:
    foil_holes = db.exec(
        select(FoilHole, GridSquare, DataCollectionGroup)
        .where(FoilHole.grid_square_id == GridSquare.id)
        .where(GridSquare.name == gsid)
        .where(GridSquare.session_id == session_id)
        .where(GridSquare.tag == DataCollectionGroup.tag)
        .where(DataCollectionGroup.id == dcgid)
    ).all()
    return [fh[0] for fh in foil_holes]


def get_foil_hole(session_id: int, fh_name: int, db) -> Dict[str, int]:
    foil_holes = db.exec(
        select(FoilHole, GridSquare)
        .where(FoilHole.name == fh_name)
        .where(FoilHole.session_id == session_id)
        .where(GridSquare.id == FoilHole.grid_square_id)
    ).all()
    return {f[1].tag: f[0].id for f in foil_holes}


def get_upstream_tiff_dirs(visit_name: str, instrument_name: str) -> List[Path]:
    tiff_dirs = []
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    for directory_name in machine_config.upstream_data_tiff_locations:
        for p in machine_config.upstream_data_directories:
            if (Path(p) / secure_filename(visit_name)).is_dir():
                processed_dir = Path(p) / secure_filename(visit_name) / directory_name
                tiff_dirs.append(processed_dir)
                break
    if not tiff_dirs:
        logger.warning(
            f"No candidate directory found for upstream download from visit {sanitise(visit_name)}"
        )
    return tiff_dirs



from unittest import mock
from unittest.mock import Mock
from urllib.parse import urlparse

import pytest

from murfey.client.tui.main import _get_visit_list
from murfey.util.models import Visit

test_get_visit_list_params_matrix = (
    ("http://0.0.0.0:8000",),
    ("http://0.0.0.0:8000/api",),
    ("http://murfey_server",),
    ("http://murfey_server/api",),
    ("http://murfey_server.com",),
)


@pytest.mark.parametrize("test_params", test_get_visit_list_params_matrix)
@mock.patch("murfey.client.tui.main.requests")
def test_get_visit_list(
    mock_request,
    test_params: tuple[str],
    mock_client_configuration,
):
    # Unpack test params and set up other params
    (server_url,) = test_params
    instrument_name = mock_client_configuration["Murfey"]["instrument_name"]

    # Construct the expected request response
    example_visits = [
        {
            "start": "1999-09-09T09:00:00",
            "end": "1999-09-11T09:00:00",
            "session_id": 123456789,
            "name": "cm12345-0",
            "beamline": "murfey",
            "proposal_title": "Commissioning Session 1",
        },
        {
            "start": "1999-09-09T09:00:00",
            "end": "1999-09-11T09:00:00",
            "session_id": 246913578,
            "name": "cm23456-1",
            "beamline": "murfey",
            "proposal_title": "Cryo-cycle 1999",
        },
    ]
    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.json.return_value = example_visits
    mock_request.get.return_value = mock_response

    # read_config() has to be patched using fixture, so has to be done in function
    with mock.patch("murfey.util.client.read_config", mock_client_configuration):
        visits = _get_visit_list(urlparse(server_url), instrument_name)

    # Check that request was sent with the correct URL
    expected_url = (
        f"{server_url}/session_control/instruments/{instrument_name}/visits_raw"
    )
    mock_request.get.assert_called_once_with(expected_url)

    # Check that expected outputs are correct (order-sensitive)
    for v, visit in enumerate(visits):
        assert (
            visit.model_dump() == Visit.model_validate(example_visits[v]).model_dump()
        )



from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import Optional

from fastapi import APIRouter, Depends
from pydantic import BaseModel
from sqlmodel import select
from werkzeug.utils import secure_filename

from murfey.server.api.auth import MurfeySessionIDInstrument as MurfeySessionID
from murfey.server.api.auth import validate_instrument_token
from murfey.server.api.file_io_shared import GainReference
from murfey.server.api.file_io_shared import process_gain as _process_gain
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise, secure_path
from murfey.util.config import get_machine_config
from murfey.util.db import Session, SessionProcessingParameters
from murfey.util.eer import num_frames

logger = getLogger("murfey.server.api.file_io_instrument")


router = APIRouter(
    prefix="/file_io/instrument",
    dependencies=[Depends(validate_instrument_token)],
    tags=["File I/O: Instrument"],
)


class SuggestedPathParameters(BaseModel):
    base_path: Path
    touch: bool = False
    extra_directory: str = ""


@router.post("/visits/{visit_name}/{session_id}/suggested_path")
def suggest_path(
    visit_name: str, session_id: int, params: SuggestedPathParameters, db=murfey_db
):
    count: Optional[int] = None
    secure_path_parts = [secure_filename(p) for p in params.base_path.parts]
    base_path = "/".join(secure_path_parts)
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if not machine_config:
        raise ValueError(
            "No machine configuration set when suggesting destination path"
        )

    # Construct the full path to where the dataset is to be saved
    check_path = machine_config.rsync_basepath / base_path

    # Check previous year to account for the year rolling over during data collection
    if not check_path.parent.exists():
        base_path_parts = base_path.split("/")
        for part in base_path_parts:
            # Find the path part corresponding to the year
            if len(part) == 4 and part.isdigit():
                year_idx = base_path_parts.index(part)
                base_path_parts[year_idx] = str(int(part) - 1)
        base_path = "/".join(base_path_parts)
        check_path_prev = check_path
        check_path = machine_config.rsync_basepath / base_path

        # If it's not in the previous year either, it's a genuine error
        if not check_path.parent.exists():
            log_message = (
                "Unable to find current visit folder under "
                f"{str(check_path_prev.parent)!r} or {str(check_path.parent)!r}"
            )
            logger.error(log_message)
            raise FileNotFoundError(log_message)

    check_path_name = check_path.name
    while check_path.exists():
        count = count + 1 if count else 2
        check_path = check_path.parent / f"{check_path_name}{count}"
    if params.touch:
        check_path.mkdir(mode=0o750)
        if params.extra_directory:
            (check_path / secure_filename(params.extra_directory)).mkdir(mode=0o750)
    return {"suggested_path": check_path.relative_to(machine_config.rsync_basepath)}


class Dest(BaseModel):
    destination: Path


@router.post("/sessions/{session_id}/make_rsyncer_destination")
def make_rsyncer_destination(session_id: int, destination: Dest, db=murfey_db):
    secure_path_parts = [secure_filename(p) for p in destination.destination.parts]
    destination_path = "/".join(secure_path_parts)
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if not machine_config:
        raise ValueError("No machine configuration set when making rsyncer destination")
    full_destination_path = machine_config.rsync_basepath / destination_path
    for parent_path in full_destination_path.parents:
        parent_path.mkdir(mode=0o750, exist_ok=True)
    return destination


@router.post("/sessions/{session_id}/process_gain")
async def process_gain(
    session_id: MurfeySessionID, gain_reference_params: GainReference, db=murfey_db
):
    result = await _process_gain(session_id, gain_reference_params, db)
    return result


class FractionationParameters(BaseModel):
    fractionation: int
    dose_per_frame: float
    num_frames: int = 0
    eer_path: Optional[str] = None
    fractionation_file_name: str = "eer_fractionation.txt"


@router.post("/visits/{visit_name}/{session_id}/eer_fractionation_file")
async def write_eer_fractionation_file(
    visit_name: str,
    session_id: int,
    fractionation_params: FractionationParameters,
    db=murfey_db,
) -> dict:
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.eer_fractionation_file_template:
        file_path = Path(
            machine_config.eer_fractionation_file_template.format(
                visit=secure_filename(visit_name),
                year=str(datetime.now().year),
            )
        ) / secure_filename(fractionation_params.fractionation_file_name)
    else:
        file_path = (
            Path(machine_config.rsync_basepath)
            / str(datetime.now().year)
            / secure_filename(visit_name)
            / machine_config.gain_directory_name
            / secure_filename(fractionation_params.fractionation_file_name)
        )

    session_parameters = db.exec(
        select(SessionProcessingParameters).where(
            SessionProcessingParameters.session_id == session_id
        )
    ).all()
    if session_parameters:
        fractionation_params.fractionation = session_parameters[0].eer_fractionation
        session_parameters[0].eer_fractionation_file = str(file_path)
        db.add(session_parameters[0])
        db.commit()

    if file_path.is_file():
        return {"eer_fractionation_file": str(file_path)}

    if fractionation_params.num_frames:
        num_eer_frames = fractionation_params.num_frames
    elif (
        fractionation_params.eer_path
        and secure_path(Path(fractionation_params.eer_path)).is_file()
    ):
        num_eer_frames = num_frames(Path(fractionation_params.eer_path))
    else:
        logger.warning(
            f"EER fractionation unable to find {secure_path(Path(fractionation_params.eer_path)) if fractionation_params.eer_path else None} "
            f"or use {int(sanitise(str(fractionation_params.num_frames)))} frames"
        )
        return {"eer_fractionation_file": None}
    with open(file_path, "w") as frac_file:
        frac_file.write(
            f"{num_eer_frames} {fractionation_params.fractionation} {fractionation_params.dose_per_frame / fractionation_params.fractionation}"
        )
    return {"eer_fractionation_file": str(file_path)}



from __future__ import annotations

import os
import socket
from functools import lru_cache
from pathlib import Path
from typing import Any, Literal, Optional

import yaml
from backports.entry_points_selectable import entry_points
from pydantic import BaseModel, ConfigDict, RootModel, ValidationInfo, field_validator
from pydantic_settings import BaseSettings


class MagnificationTable(RootModel[dict[int, float]]):
    pass


CALIBRATIONS_VALIDATION_SCHEMAS = {
    "magnification": MagnificationTable,
}


class MachineConfig(BaseModel):  # type: ignore
    """
    Keys that describe the type of workflow conducted on the client side, and how
    Murfey will handle its data transfer and processing
    """

    # General info --------------------------------------------------------------------
    display_name: str = ""
    instrument_name: str = ""
    image_path: Optional[Path] = None
    machine_override: str = ""

    # Hardware and software -----------------------------------------------------------
    camera: str = "FALCON"
    superres: bool = False
    calibrations: dict[str, Any]
    acquisition_software: list[str]
    software_versions: dict[str, str] = {}
    software_settings_output_directories: dict[str, list[str]] = {}
    data_required_substrings: dict[str, dict[str, list[str]]] = {}

    # Client side directory setup -----------------------------------------------------
    data_directories: list[Path]
    create_directories: list[str] = ["atlas"]
    analyse_created_directories: list[str] = []
    gain_reference_directory: Optional[Path] = None
    eer_fractionation_file_template: str = ""

    # Data transfer setup -------------------------------------------------------------
    # Rsync setup
    data_transfer_enabled: bool = True
    rsync_url: str = ""
    rsync_module: str = ""
    rsync_basepath: Path
    allow_removal: bool = False

    # Upstream data download setup
    upstream_data_directories: list[Path] = []  # Previous sessions
    upstream_data_download_directory: Optional[Path] = None  # Set by microscope config
    upstream_data_tiff_locations: list[str] = ["processed"]  # Location of CLEM TIFFs

    # Data processing setup -----------------------------------------------------------
    # General processing setup
    processing_enabled: bool = True
    process_by_default: bool = True
    gain_directory_name: str = "processing"
    process_multiple_datasets: bool = True
    processed_directory_name: str = "processed"
    processed_extra_directory: str = ""
    recipes: dict[str, str] = {
        "em-spa-bfactor": "em-spa-bfactor",
        "em-spa-class2d": "em-spa-class2d",
        "em-spa-class3d": "em-spa-class3d",
        "em-spa-preprocess": "em-spa-preprocess",
        "em-spa-refine": "em-spa-refine",
        "em-tomo-preprocess": "em-tomo-preprocess",
        "em-tomo-align": "em-tomo-align",
    }

    # Particle picking setup
    default_model: Path
    picking_model_search_directory: str = "processing"
    initial_model_search_directory: str = "processing/initial_model"

    # Data analysis plugins
    external_executables: dict[str, str] = {}
    external_executables_eer: dict[str, str] = {}
    external_environment: dict[str, str] = {}
    plugin_packages: dict[str, Path] = {}

    # Server and network setup --------------------------------------------------------
    # Configurations and URLs
    security_configuration_path: Optional[Path] = None
    murfey_url: str = "http://localhost:8000"
    frontend_url: str = "http://localhost:3000"
    instrument_server_url: str = "http://localhost:8001"

    # Messaging queues
    failure_queue: str = ""
    node_creator_queue: str = "node_creator"
    notifications_queue: str = "pato_notification"

    # Pydantic BaseModel settings
    model_config = ConfigDict(extra="allow")

    @field_validator("calibrations", mode="before")
    @classmethod
    def validate_calibration_data(
        cls, v: dict[str, dict[Any, Any]]
    ) -> dict[str, dict[Any, Any]]:
        # Pass the calibration dictionaries through their matching Pydantic models, if any are set
        if isinstance(v, dict):
            validated = {}
            for (
                key,
                value,
            ) in v.items():
                model_cls = CALIBRATIONS_VALIDATION_SCHEMAS.get(key)
                if model_cls:
                    try:
                        # Validate and store as a dict object with the corrected types
                        validated[key] = model_cls.model_validate(value).root
                    except Exception as e:
                        raise ValueError(f"Validation failed for key '{key}': {e}")
                else:
                    validated[key] = value
            return validated
        # Let it validate and fail as-is
        return v

    @field_validator("software_versions", mode="before")
    @classmethod
    def validate_software_versions(cls, v: dict[str, Any]) -> dict[str, str]:
        # Software versions should be numerical strings, even if they appear int- or float-like
        if isinstance(v, dict):
            validated = {key: str(value) for key, value in v.items()}
            return validated
        # Let it validate and fail as-is
        return v


def from_file(config_file_path: Path, instrument: str = "") -> dict[str, MachineConfig]:
    with open(config_file_path, "r") as config_stream:
        config = yaml.safe_load(config_stream)
    return {
        i: MachineConfig(**config[i])
        for i in config.keys()
        if not instrument or i == instrument
    }


class Security(BaseModel):
    # Murfey database settings
    murfey_db_credentials: Path
    crypto_key: str
    sqlalchemy_pooling: bool = True

    # ISPyB settings
    ispyb_credentials: Optional[Path] = None

    # Murfey server connection settings
    auth_url: str = ""
    auth_type: Literal["password", "cookie"] = "password"
    auth_algorithm: str = ""
    auth_key: str = ""
    cookie_key: str = ""
    instrument_auth_url: str = ""
    instrument_auth_type: Literal["token", ""] = "token"
    allow_user_token: bool = False  # TUI 'user' token support
    session_validation: str = ""
    session_token_timeout: Optional[int] = None
    allow_origins: list[str] = ["*"]

    # RabbitMQ settings
    rabbitmq_credentials: Path
    feedback_queue: str = "murfey_feedback"

    # Graylog settings
    graylog_host: str = ""
    graylog_port: Optional[int] = None

    model_config = ConfigDict()

    @field_validator("graylog_port")
    def check_port_present_if_host_is(
        cls, v: Optional[int], info: ValidationInfo, **kwargs
    ) -> Optional[int]:
        if info.data.get("graylog_host") and v is None:
            raise ValueError("The Graylog port must be set if the Graylog host is")
        return v


def security_from_file(config_file_path: Path) -> Security:
    with open(config_file_path, "r") as config_stream:
        config = yaml.safe_load(config_stream)
    return Security(**config)


class Settings(BaseSettings):
    murfey_machine_configuration: str = ""
    murfey_security_configuration: str = ""


settings = Settings()


@lru_cache()
def get_hostname():
    return socket.gethostname()


def get_microscope(machine_config: MachineConfig | None = None) -> str:
    if machine_config:
        microscope_name = machine_config.machine_override or os.getenv("BEAMLINE", "")
    else:
        microscope_name = os.getenv("BEAMLINE", "")
    return microscope_name


@lru_cache(maxsize=1)
def get_security_config() -> Security:
    if settings.murfey_security_configuration:
        return security_from_file(Path(settings.murfey_security_configuration))
    if settings.murfey_machine_configuration and os.getenv("BEAMLINE"):
        machine_config = get_machine_config(instrument_name=os.getenv("BEAMLINE"))[
            os.getenv("BEAMLINE", "")
        ]
        if machine_config.security_configuration_path:
            return security_from_file(machine_config.security_configuration_path)
    return Security(
        rabbitmq_credentials="",
        session_validation="",
        murfey_db_credentials="",
        crypto_key="",
        auth_key="",
        auth_algorithm="",
        sqlalchemy_pooling=True,
    )


@lru_cache(maxsize=1)
def get_machine_config(instrument_name: str = "") -> dict[str, MachineConfig]:
    machine_config = {
        "": MachineConfig(
            acquisition_software=[],
            calibrations={},
            data_directories=[],
            rsync_basepath=Path("dls/tmp"),
            murfey_db_credentials="",
            default_model="/tmp/weights.h5",
        )
    }
    if settings.murfey_machine_configuration:
        microscope = instrument_name
        machine_config = from_file(
            Path(settings.murfey_machine_configuration), microscope
        )
    return machine_config


def get_extended_machine_config(
    extension_name: str, instrument_name: str = ""
) -> Optional[BaseModel]:
    machine_config = get_machine_config(instrument_name=instrument_name).get(
        instrument_name or get_microscope()
    )
    if not machine_config:
        return None
    model = entry_points().select(group="murfey.config", name=extension_name)[0].load()
    data = getattr(machine_config, extension_name, {})
    return model(**data)



from pathlib import Path
from unittest.mock import MagicMock

import pytest

from murfey.server.ispyb import TransportManager
from murfey.workflows.clem.process_raw_tiffs import zocalo_cluster_request

# Set up variables
session_id = 0
instrument_name = "clem"
root_folder = "images"
visit_name = "cm12345-6"
area_name = "test_area"
feedback_queue = "murfey_feedback"

# Properties for TIFF images
num_z = 5
num_c = 3


@pytest.fixture
def raw_dir(tmp_path: Path):
    raw_dir = tmp_path / visit_name / root_folder
    raw_dir.mkdir(parents=True, exist_ok=True)
    return raw_dir


@pytest.fixture
def tiff_list(raw_dir: Path):
    (raw_dir / area_name).mkdir(parents=True, exist_ok=True)
    tiff_list = [
        raw_dir / area_name / f"test_series--Z{str(z).zfill(2)}--C{str(c).zfill(2)}.tif"
        for z in range(num_z)
        for c in range(num_c)
    ]
    for file in tiff_list:
        if not file.exists():
            file.touch()
    return tiff_list


@pytest.fixture
def metadata(raw_dir: Path):
    (raw_dir / area_name / "Metadata").mkdir(parents=True, exist_ok=True)
    metadata = raw_dir / area_name / "Metadata" / "test_series.xlif"
    if not metadata.exists():
        metadata.touch()
    return metadata


def test_zocalo_cluster_request(
    tiff_list: list[Path],
    metadata: Path,
    raw_dir: Path,
):

    # Create a mock tranpsort object
    mock_transport = MagicMock(spec=TransportManager)
    mock_transport.feedback_queue = feedback_queue

    # Run the function with the listed parameters
    zocalo_cluster_request(
        tiff_list=tiff_list,
        root_folder=root_folder,
        session_id=session_id,
        instrument_name=instrument_name,
        metadata=metadata,
        messenger=mock_transport,
    )

    # Construct the recipe that we expect to send
    job_name = "--".join(
        [
            p.replace(" ", "_") if " " in p else p
            for p in (
                tiff_list[0].parent.relative_to(raw_dir)
                / tiff_list[0].stem.split("--")[0]
            ).parts
        ]
    )
    sent_recipe = {
        "recipes": ["clem-tiff-to-stack"],
        "parameters": {
            # Job parameters
            "tiff_list": "null",
            "tiff_file": f"{str(tiff_list[0])}",
            "root_folder": root_folder,
            "metadata": f"{str(metadata)}",
            # Other recipe parameters
            "session_dir": f"{str(raw_dir.parent)}",
            "session_id": session_id,
            "job_name": job_name,
            "feedback_queue": feedback_queue,
        },
    }

    # Check that it sends the expected recipe
    mock_transport.send.assert_called_once_with(
        queue="processing_recipe",
        message=sent_recipe,
        new_connection=True,
    )



from __future__ import annotations

import logging
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Any, Callable, Optional
from uuid import uuid4

from werkzeug.utils import secure_filename

logger = logging.getLogger("murfey.util")


def sanitise(in_string: str) -> str:
    return in_string.replace("\r\n", "").replace("\n", "")


def sanitise_path(in_path: Path) -> Path:
    return Path("/".join(secure_filename(p) for p in in_path.parts))


def sanitise_nonpath(in_string: str) -> str:
    for c in ("\r\n", "\n", "/", "\\", ":", ";"):
        in_string = in_string.replace(c, "")
    return in_string


def secure_path(in_path: Path, keep_spaces: bool = False) -> Path:
    if keep_spaces:
        secured_parts = []
        for p, part in enumerate(in_path.parts):
            if " " in part:
                secured_parts.append(part)
            elif ":" in part and not p:
                secured_parts.append(secure_filename(part) + ":")
            else:
                secured_parts.append(secure_filename(part))
    else:
        secured_parts = [
            (
                secure_filename(part) + ":"
                if p == 0 and ":" in part
                else secure_filename(part)
            )
            for p, part in enumerate(in_path.parts)
        ]
    return Path("/".join(secured_parts))


def posix_path(path: Path) -> str:
    """
    Converts a Windows-style path into a Posix one. Used primarily when running
    subproceses in bash terminals on Windows devices, which can only accept
    Posix paths.

    Returns it as a string because this path won't be recognised as an existing
    path when converted to a Path object.
    """
    path_parts = list(path.parts)
    # Check if it's a Windows-style path and converts it to a Posix one
    #   e.g.: C:\Users\user -> /c/Users/user
    if path_parts[0].endswith((":/", ":\\")):
        path_parts[0] = "/" + path_parts[0].strip(":/\\").lower()
        posix_path = "/".join(path_parts)
        return posix_path
    return str(path)


class Processor:
    def __init__(self, name: Optional[str] = None):
        self._in: Queue = Queue()
        self._out: Queue = Queue()
        self._previous: Optional[Processor] = None
        self.thread: Optional[Thread] = None
        self.name = name or str(uuid4())[:8]

    def __rshift__(self, other: Processor):
        self.point_to(other)

    def point_to(self, other: Processor):
        if isinstance(other, Processor):
            other._in = self._out
            other._previous = self

    def process(self, in_thread: bool = False, thread_name: str = "", **kwargs):
        if in_thread:
            self.thread = Thread(
                target=self._process,
                kwargs=kwargs,
                name=thread_name or self.name,
            )
            self.thread.start()
        else:
            self._process(**kwargs)

    def _process(self, **kwargs):
        pass

    def wait(self):
        if self.thread:
            self.thread.join()


class LogFilter(logging.Filter):
    """A filter to limit messages going to Graylog"""

    def __repr__(self):
        return "<murfey.server.LogFilter>"

    def __init__(self):
        super().__init__()
        self._filter_levels = {
            "murfey": logging.DEBUG,
            "ispyb": logging.DEBUG,
            "zocalo": logging.DEBUG,
            "uvicorn": logging.INFO,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.INFO,
        }

    @staticmethod
    def install() -> LogFilter:
        logfilter = LogFilter()
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            handler.addFilter(logfilter)
        return logfilter

    def filter(self, record: logging.LogRecord) -> bool:
        logger_name = record.name
        while True:
            if logger_name in self._filter_levels:
                return record.levelno >= self._filter_levels[logger_name]
            if "." not in logger_name:
                return False
            logger_name = logger_name.rsplit(".", maxsplit=1)[0]


def safe_run(
    func: Callable,
    args: list | tuple = [],
    kwargs: dict[str, Any] = {},
    label: str = "",
):
    """
    A wrapper to encase individual functions in try-except blocks so that a warning
    is raised if the function fails, but the process continues as normal otherwise.
    """
    try:
        return func(*args, **kwargs)
    except Exception:
        logger.warning(
            f"Function {func.__name__!r} failed to run for object {label!r}",
            exc_info=True,
        )
        return None



"""
Contains functions for analysing the various types of files hauled by Murfey, and
assigning to them the correct contexts (CLEM, SPA, tomography, etc.) for processing
on the server side.

Individual contexts can be found in murfey.client.contexts.
"""

from __future__ import annotations

import logging
import queue
import threading
from pathlib import Path
from typing import Type

from murfey.client.context import Context
from murfey.client.contexts.clem import CLEMContext
from murfey.client.contexts.spa import SPAModularContext
from murfey.client.contexts.spa_metadata import SPAMetadataContext
from murfey.client.contexts.tomo import TomographyContext
from murfey.client.contexts.tomo_metadata import TomographyMetadataContext
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.rsync import RSyncerUpdate, TransferResult
from murfey.util.client import Observer, get_machine_config_client
from murfey.util.mdoc import get_block
from murfey.util.models import ProcessingParametersSPA, ProcessingParametersTomo

logger = logging.getLogger("murfey.client.analyser")


class Analyser(Observer):
    def __init__(
        self,
        basepath_local: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        force_mdoc_metadata: bool = False,
        limited: bool = False,
    ):
        super().__init__()
        self._basepath = basepath_local.absolute()
        self._limited = limited
        self._experiment_type = ""
        self._acquisition_software = ""
        self._extension: str = ""
        self._unseen_xml: list = []
        self._context: Context | None = None
        self._batch_store: dict = {}
        self._environment = environment
        self._force_mdoc_metadata = force_mdoc_metadata
        self.parameters_model: (
            Type[ProcessingParametersSPA] | Type[ProcessingParametersTomo] | None
        ) = None

        self.queue: queue.Queue = queue.Queue()
        self.thread = threading.Thread(name="Analyser", target=self._analyse)
        self._stopping = False
        self._halt_thread = False
        self._murfey_config = (
            get_machine_config_client(
                str(environment.url.geturl()),
                instrument_name=environment.instrument_name,
                demo=environment.demo,
            )
            if environment
            else {}
        )

    def __repr__(self) -> str:
        return f"<Analyser ({self._basepath})>"

    def _find_extension(self, file_path: Path) -> bool:
        """
        Identifies the file extension and stores that information in the class.
        """
        if "atlas" in file_path.parts:
            self._extension = file_path.suffix
            return True

        if (
            required_substrings := self._murfey_config.get(
                "data_required_substrings", {}
            )
            .get(self._acquisition_software, {})
            .get(file_path.suffix)
        ):
            if not any(r in file_path.name for r in required_substrings):
                return False

        # Checks for MRC, TIFF, TIF, and EER files
        if file_path.suffix in (".mrc", ".tiff", ".tif", ".eer"):
            if not self._extension:
                logger.info(f"File extension determined: {file_path.suffix}")
                self._extension = file_path.suffix
            elif self._extension != file_path.suffix:
                logger.info(f"File extension re-evaluated: {file_path.suffix}")
                self._extension = file_path.suffix
            return True
        # If we see an .mdoc file first, use that to determine the file extensions
        elif file_path.suffix == ".mdoc":
            with open(file_path, "r") as md:
                md.seek(0)
                mdoc_data_block = get_block(md)
            if subframe_path := mdoc_data_block.get("SubFramePath"):
                self._extension = Path(subframe_path).suffix
                return True
        # Check for LIF files separately
        elif file_path.suffix == ".lif":
            self._extension = file_path.suffix
            return True
        return False

    def _find_context(self, file_path: Path) -> bool:
        """
        Using various conditionals, identifies what workflow the file is part of, and
        assigns the correct context class to that batch of rsync files for subsequent
        stages of processing. Actions to take for individual files will be determined
        in the Context classes themselves.
        """
        logger.debug(f"Finding context using file {str(file_path)!r}")

        # CLEM workflow checks
        # Look for LIF and XLIF files
        if file_path.suffix in (".lif", ".xlif"):
            self._context = CLEMContext("leica", self._basepath)
            return True
        # Look for TIFF files associated with CLEM workflow
        # Leica's autosave mode seems to name the TIFFs in the format
        # PostionXX--ZXX--CXX.tif
        if all(
            pattern in file_path.name for pattern in ("--Z", "--C")
        ) and file_path.suffix in (".tiff", ".tif"):
            self._context = CLEMContext("leica", self._basepath)
            return True

        # Tomography and SPA workflow checks
        if "atlas" in file_path.parts:
            self._context = SPAMetadataContext("epu", self._basepath)
            return True

        if "Metadata" in file_path.parts or file_path.name == "EpuSession.dm":
            self._context = SPAMetadataContext("epu", self._basepath)
            return True
        elif (
            "Batch" in file_path.parts
            or "SearchMaps" in file_path.parts
            or "Thumbnails" in file_path.parts
            or file_path.name == "Session.dm"
        ):
            self._context = TomographyMetadataContext("tomo", self._basepath)
            return True

        split_file_stem = file_path.stem.split("_")
        if split_file_stem:
            if split_file_stem[-1] == "gain":
                return False

            # Files starting with "FoilHole" belong to the SPA workflow
            if split_file_stem[0].startswith("FoilHole") and split_file_stem[-1] in [
                "Fractions",
                "fractions",
                "EER",
            ]:
                if not self._context:
                    logger.info("Acquisition software: EPU")
                    self._context = SPAModularContext("epu", self._basepath)
                self.parameters_model = ProcessingParametersSPA
                return True

            # Files starting with "Position" belong to the standard tomography workflow
            # NOTE: not completely reliable, mdocs can be in tomography metadata as well
            if (
                split_file_stem[0] == "Position"
                or "[" in file_path.name
                or split_file_stem[-1] in ["Fractions", "fractions", "EER"]
                or file_path.suffix == ".mdoc"
            ):
                if not self._context:
                    logger.info("Acquisition software: tomo")
                    self._context = TomographyContext("tomo", self._basepath)
                    self.parameters_model = ProcessingParametersTomo
                return True
        return False

    def post_transfer(self, transferred_file: Path):
        try:
            if self._context:
                self._context.post_transfer(
                    transferred_file, environment=self._environment
                )
        except Exception as e:
            logger.error(
                f"An exception was encountered post transfer: {e}", exc_info=True
            )

    def _analyse(self):
        logger.info("Analyser thread started")
        mdoc_for_reading = None
        while not self._halt_thread:
            transferred_file = self.queue.get()
            transferred_file = (
                Path(transferred_file)
                if isinstance(transferred_file, str)
                else transferred_file
            )
            if not transferred_file:
                self._halt_thread = True
                continue
            if self._limited:
                if (
                    "Metadata" in transferred_file.parts
                    or transferred_file.name == "EpuSession.dm"
                    and not self._context
                ):
                    self._context = SPAMetadataContext("epu", self._basepath)
                elif (
                    "Batch" in transferred_file.parts
                    or "SearchMaps" in transferred_file.parts
                    or transferred_file.name == "Session.dm"
                    and not self._context
                ):
                    self._context = TomographyMetadataContext("tomo", self._basepath)
                self.post_transfer(transferred_file)
            else:
                dc_metadata = {}
                if (
                    self._force_mdoc_metadata
                    and transferred_file.suffix == ".mdoc"
                    or mdoc_for_reading
                ):
                    if self._context:
                        try:
                            dc_metadata = self._context.gather_metadata(
                                mdoc_for_reading or transferred_file,
                                environment=self._environment,
                            )
                        except KeyError as e:
                            logger.error(
                                f"Metadata gathering failed with a key error for key: {e.args[0]}"
                            )
                            raise e
                        if not dc_metadata:
                            mdoc_for_reading = None
                    elif transferred_file.suffix == ".mdoc":
                        mdoc_for_reading = transferred_file
                if not self._context:
                    valid_extension = self._find_extension(transferred_file)
                    if not valid_extension:
                        logger.error(f"No extension found for {transferred_file}")
                        continue
                    found = self._find_context(transferred_file)
                    if not found:
                        logger.debug(
                            f"Couldn't find context for {str(transferred_file)!r}"
                        )
                        self.queue.task_done()
                        continue
                    elif self._extension:
                        logger.info(
                            f"Context found successfully for {transferred_file}"
                        )
                        try:
                            self._context.post_first_transfer(
                                transferred_file,
                                environment=self._environment,
                            )
                        except Exception as e:
                            logger.error(f"Exception encountered: {e}")
                        if "atlas" not in transferred_file.parts:
                            if not dc_metadata:
                                try:
                                    dc_metadata = self._context.gather_metadata(
                                        self._xml_file(transferred_file),
                                        environment=self._environment,
                                    )
                                except NotImplementedError:
                                    dc_metadata = {}
                                except KeyError as e:
                                    logger.error(
                                        f"Metadata gathering failed with a key error for key: {e.args[0]}"
                                    )
                                    raise e
                                except ValueError as e:
                                    logger.error(
                                        f"Metadata gathering failed with a value error: {e}"
                                    )
                            if not dc_metadata or not self._force_mdoc_metadata:
                                self._unseen_xml.append(transferred_file)
                            else:
                                self._unseen_xml = []
                                if dc_metadata.get("file_extension"):
                                    self._extension = dc_metadata["file_extension"]
                                else:
                                    dc_metadata["file_extension"] = self._extension
                                dc_metadata["acquisition_software"] = (
                                    self._context._acquisition_software
                                )
                                self.notify(
                                    {
                                        "form": dc_metadata,
                                    }
                                )

                # If a file with a CLEM context is identified, immediately post it
                elif isinstance(self._context, CLEMContext):
                    logger.debug(
                        f"File {transferred_file.name!r} will be processed as part of CLEM workflow"
                    )
                    self.post_transfer(transferred_file)

                # Handle files with tomography and SPA context differently
                elif not self._extension or self._unseen_xml:
                    valid_extension = self._find_extension(transferred_file)
                    if not valid_extension:
                        logger.error(f"No extension found for {transferred_file}")
                        continue
                    if self._extension:
                        logger.info(
                            f"Extension found successfully for {transferred_file}"
                        )
                        try:
                            self._context.post_first_transfer(
                                transferred_file,
                                environment=self._environment,
                            )
                        except Exception as e:
                            logger.error(f"Exception encountered: {e}")
                        if "atlas" not in transferred_file.parts:
                            if not dc_metadata:
                                try:
                                    dc_metadata = self._context.gather_metadata(
                                        mdoc_for_reading
                                        or self._xml_file(transferred_file),
                                        environment=self._environment,
                                    )
                                except KeyError as e:
                                    logger.error(
                                        f"Metadata gathering failed with a key error for key: {e.args[0]}"
                                    )
                                    raise e
                            if not dc_metadata or not self._force_mdoc_metadata:
                                mdoc_for_reading = None
                                self._unseen_xml.append(transferred_file)
                            if dc_metadata:
                                self._unseen_xml = []
                                if dc_metadata.get("file_extension"):
                                    self._extension = dc_metadata["file_extension"]
                                else:
                                    dc_metadata["file_extension"] = self._extension
                                dc_metadata["acquisition_software"] = (
                                    self._context._acquisition_software
                                )
                                self.notify(
                                    {
                                        "form": dc_metadata,
                                    }
                                )
                elif isinstance(
                    self._context,
                    (
                        SPAModularContext,
                        SPAMetadataContext,
                        TomographyContext,
                        TomographyMetadataContext,
                    ),
                ):
                    context = str(self._context).split(" ")[0].split(".")[-1]
                    logger.debug(
                        f"Transferring file {str(transferred_file)} with context {context!r}"
                    )
                    self.post_transfer(transferred_file)
            self.queue.task_done()
        logger.debug("Analyer thread has stopped analysing incoming files")
        self.notify(final=True)

    def _xml_file(self, data_file: Path) -> Path:
        if not self._environment:
            return data_file.with_suffix(".xml")
        file_name = f"{'_'.join(p for p in data_file.stem.split('_')[:-1])}.xml"
        data_directories = self._murfey_config.get("data_directories", [])
        for dd in data_directories:
            if str(data_file).startswith(dd):
                base_dir = Path(dd).absolute()
                mid_dir = data_file.relative_to(base_dir).parent
                break
        else:
            return data_file.with_suffix(".xml")
        return base_dir / self._environment.visit / mid_dir / file_name

    def enqueue(self, rsyncer: RSyncerUpdate):
        if not self._stopping and rsyncer.outcome == TransferResult.SUCCESS:
            absolute_path = (self._basepath / rsyncer.file_path).absolute()
            self.queue.put(absolute_path)

    def start(self):
        if self.thread.is_alive():
            raise RuntimeError("Analyser already running")
        if self._stopping:
            raise RuntimeError("Analyser has already stopped")
        logger.info(f"Analyser thread starting for {self}")
        self.thread.start()

    def request_stop(self):
        self._stopping = True
        self._halt_thread = True

    def is_safe_to_stop(self):
        """
        Checks that the analyser thread is safe to stop
        """
        return self._stopping and self._halt_thread and not self.queue.qsize()

    def stop(self):
        logger.debug("Analyser thread stop requested")
        self._stopping = True
        self._halt_thread = True
        try:
            if self.thread.is_alive():
                self.queue.put(None)
                self.thread.join()
        except Exception as e:
            logger.error(
                f"Exception encountered while stopping Analyser: {e}",
                exc_info=True,
            )
        logger.debug("Analyser thread stop completed")



def midpoint(angles: list[float]) -> int:
    """
    Utility function to calculate the midpoint of the angles used in a tilt series.
    Used primarily in the tomography workflow.
    """
    if not angles:
        return 0
    if len(angles) <= 2:
        return round(angles[0])
    sorted_angles = sorted(angles)
    return round(
        sorted_angles[len(sorted_angles) // 2]
        if sorted_angles[len(sorted_angles) // 2]
        and sorted_angles[len(sorted_angles) // 2 + 1]
        else 0
    )



from __future__ import annotations

from importlib.resources import files
from typing import TYPE_CHECKING

# Classes are only imported for type checking purposes
if TYPE_CHECKING:
    from uvicorn import Server

    from murfey.server.ispyb import TransportManager

_running_server: Server | None = None
_transport_object: TransportManager | None = None
template_files = files("murfey") / "templates"



from datetime import datetime
from functools import lru_cache
from pathlib import Path
from typing import Literal, Optional

from pydantic import BaseModel
from werkzeug.utils import secure_filename

from murfey.util.config import MachineConfig, get_machine_config


def motion_corrected_mrc(
    input_movie: Path, visit_name: str, machine_config: MachineConfig
):
    parts = [secure_filename(p) for p in input_movie.parts]
    visit_idx = parts.index(visit_name)
    core = Path("/") / Path(*parts[: visit_idx + 1])
    ppath = Path("/") / Path(*parts)
    if machine_config.process_multiple_datasets:
        sub_dataset = ppath.relative_to(core).parts[0]
    else:
        sub_dataset = ""
    extra_path = machine_config.processed_extra_directory
    mrc_out = (
        core
        / machine_config.processed_directory_name
        / sub_dataset
        / extra_path
        / "MotionCorr"
        / "job002"
        / "Movies"
        / ppath.parent.relative_to(core / sub_dataset)
        / str(ppath.stem + "_motion_corrected.mrc")
    )
    return Path("/".join(secure_filename(p) for p in mrc_out.parts))


@lru_cache(maxsize=5)
def cryolo_model_path(visit: str, instrument_name: str) -> Path:
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.picking_model_search_directory:
        visit_directory = (
            machine_config.rsync_basepath / str(datetime.now().year) / visit
        )
        possible_models = list(
            (visit_directory / machine_config.picking_model_search_directory).glob(
                "*.h5"
            )
        )
        if possible_models:
            return sorted(possible_models, key=lambda x: x.stat().st_ctime)[-1]
    return machine_config.default_model


class CLEMAlignAndMergeParameters(BaseModel):
    crop_to_n_frames: Optional[int] = 50
    align_self: Literal["enabled", ""] = "enabled"
    flatten: Literal["mean", "min", "max", ""] = "mean"
    align_across: Literal["enabled", ""] = "enabled"


default_clem_align_and_merge_parameters = CLEMAlignAndMergeParameters()


class SPAParameters(BaseModel):
    nr_iter_2d: int = 25
    nr_iter_3d: int = 25
    nr_iter_ini_model: int = 100
    batch_size_2d: int = 50000
    nr_classes_2d: int = 50
    nr_classes_3d: int = 4
    downscale: bool = True
    do_icebreaker_jobs: bool = True
    fraction_of_classes_to_remove_2d: float = 0.7
    nr_picks_before_diameter: int = 10000
    bfactor_min_particles: int = 2000


default_spa_parameters = SPAParameters()



from __future__ import annotations

import itertools
import logging
from itertools import count
from pathlib import Path
from threading import RLock
from typing import Dict, List, NamedTuple, Optional
from urllib.parse import ParseResult

from pydantic import BaseModel, ConfigDict

from murfey.client.watchdir import DirWatcher

logger = logging.getLogger("murfey.client.instance_environment")

MurfeyID = count(1)
MovieID = count(1)


class MovieTracker(NamedTuple):
    movie_number: int
    motion_correction_uuid: int


class SampleInfo(NamedTuple):
    atlas: Path
    sample: int


global_env_lock = RLock()


class MurfeyInstanceEnvironment(BaseModel):
    url: ParseResult
    client_id: int
    instrument_name: str
    software_versions: Dict[str, str] = {}
    sources: List[Path] = []
    default_destinations: Dict[Path, str] = {}
    destination_registry: Dict[str, str] = {}
    watchers: Dict[Path, DirWatcher] = {}
    demo: bool = False
    movies: Dict[Path, MovieTracker] = {}
    movie_tilt_pair: Dict[Path, str] = {}
    tilt_angles: Dict[str, List[List[str]]] = {}
    movie_counters: Dict[str, itertools.count] = {}
    visit: str = ""
    processing_only_mode: bool = False
    dose_per_frame: Optional[float] = None
    gain_ref: Optional[str] = None
    symmetry: Optional[str] = None
    eer_fractionation: Optional[int] = None
    superres: bool = False
    murfey_session: Optional[int] = None
    samples: Dict[Path, SampleInfo] = {}
    rsync_url: str = ""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def clear(self):
        self.sources = []
        self.default_destinations = {}
        for w in self.watchers.values():
            w.stop()
        self.watchers = {}
        self.movies = {}
        self.movie_tilt_pair = {}
        self.tilt_angles = {}
        self.visit = ""
        self.dose_per_frame = None
        self.gain_ref = None
        self.symmetry = None
        self.eer_fractionation = None



import secrets
import string

from cryptography.fernet import Fernet

from murfey.util.config import get_security_config


def run():
    security_config = get_security_config()
    f = Fernet(security_config.crypto_key.encode("ascii"))
    alphabet = string.ascii_letters + string.digits
    password = "".join(secrets.choice(alphabet) for i in range(32))
    print(f.encrypt(password.encode("ascii")).decode())



from __future__ import annotations

import asyncio
import datetime
import logging
from pathlib import Path
from typing import Annotated, Any, List, Optional
from urllib.parse import quote

import aiohttp
from fastapi import APIRouter, Depends
from pydantic import BaseModel
from sqlmodel import select
from werkzeug.utils import secure_filename

from murfey.server.api.auth import MurfeyInstrumentNameFrontend as MurfeyInstrumentName
from murfey.server.api.auth import MurfeySessionIDFrontend as MurfeySessionID
from murfey.server.api.auth import (
    create_access_token,
    instrument_server_tokens,
    oauth2_scheme,
    validate_token,
)
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise, secure_path
from murfey.util.api import url_path_for
from murfey.util.config import get_machine_config
from murfey.util.db import RsyncInstance, Session, SessionProcessingParameters
from murfey.util.models import File, MultigridWatcherSetup

# Create APIRouter class object
router = APIRouter(
    prefix="/instrument_server",
    dependencies=[Depends(validate_token)],
    tags=["Instrument Server"],
)

log = logging.getLogger("murfey.server.instrument")

lock = asyncio.Lock()


@router.post(
    "/instruments/{instrument_name}/sessions/{session_id}/activate_instrument_server"
)
async def activate_instrument_server_for_session(
    instrument_name: MurfeyInstrumentName,
    session_id: int,
    token_in: Annotated[str, Depends(oauth2_scheme)],
    db=murfey_db,
):
    log.info(
        f"Activating instrument server for session {int(sanitise(str(session_id)))}"
    )
    if not session_id > 0:
        log.warning("Invalid session ID")
        return False
    visit_name = db.exec(select(Session).where(Session.id == session_id)).one().visit
    timestamp = datetime.datetime.now().timestamp()
    token = create_access_token(
        {"timestamp": timestamp, "session": session_id, "visit": visit_name},
        token=token_in,
    )
    async with lock:
        instrument_server_tokens[session_id] = {}
        machine_config = get_machine_config(instrument_name=instrument_name)[
            instrument_name
        ]
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'token_handshake_for_session', session_id=session_id)}",
                json={"access_token": token, "token_type": "bearer"},
            ) as response:
                success = response.status == 200
                instrument_server_token = await response.json()
                instrument_server_tokens[session_id] = instrument_server_token
    if success:
        log.info("Handshake successful")
    else:
        log.warning("Handshake unsuccessful")
    return success


@router.get("/instruments/{instrument_name}/sessions/{session_id}/active")
async def check_if_session_is_active(
    instrument_name: MurfeyInstrumentName, session_id: int
):
    if instrument_server_tokens.get(session_id) is None:
        return {"active": False}
    async with lock:
        async with aiohttp.ClientSession() as clientsession:
            machine_config = get_machine_config(instrument_name=instrument_name)[
                instrument_name
            ]
            async with clientsession.get(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'check_token', session_id=session_id)}",
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as response:
                return {"active": response.status == 200}


@router.get("/sessions/{session_id}/multigrid_controller/status")
async def check_multigrid_controller_status(session_id: MurfeySessionID, db=murfey_db):
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    instrument_name = session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        log.debug(
            f"Submitting request to inspect multigrid controller for session {session_id}"
        )
        # Treat it as absent if the server-side has no stored token for the session
        if (
            instrument_server_tokens.get(session_id, {}).get("access_token", None)
            is None
        ):
            return {"exists": False}
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.get(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'check_multigrid_controller_status', session_id=session_id)}",
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data: dict[str, Any] = await resp.json()
    else:
        data = {"detail": "No instrument server URL found"}
    log.debug(f"Received response: {data}")
    return data


@router.post("/sessions/{session_id}/multigrid_watcher")
async def setup_multigrid_watcher(
    session_id: MurfeySessionID, watcher_spec: MultigridWatcherSetup, db=murfey_db
):
    data = {}
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    instrument_name = session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        session = db.exec(select(Session).where(Session.id == session_id)).one()
        visit = session.visit
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'setup_multigrid_watcher', session_id=session_id)}",
                json={
                    "source": str(secure_path(watcher_spec.source / visit)),
                    "visit": visit,
                    "label": visit,
                    "instrument_name": instrument_name,
                    "skip_existing_processing": watcher_spec.skip_existing_processing,
                    "destination_overrides": {
                        str(k): v for k, v in watcher_spec.destination_overrides.items()
                    },
                    "rsync_restarts": watcher_spec.rsync_restarts,
                    "visit_end_time": (
                        str(session.visit_end_time) if session.visit_end_time else None
                    ),
                },
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


@router.post("/sessions/{session_id}/start_multigrid_watcher")
async def start_multigrid_watcher(session_id: MurfeySessionID, db=murfey_db):
    data = {}
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    process = session.process
    instrument_name = session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        log.debug(
            f"Submitting request to start multigrid watcher for session {session_id} "
            f"with processing {('enabled' if process else 'disabled')}"
        )
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'start_multigrid_watcher', session_id=session_id)}?process={'true' if process else 'false'}",
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    log.debug(f"Received response: {data}")
    return data


@router.post("/sessions/{session_id}/multigrid_controller/visit_end_time")
async def update_visit_end_time(
    session_id: MurfeySessionID, end_time: datetime.datetime, db=murfey_db
):
    # Load data for session
    session_entry = db.exec(select(Session).where(Session.id == session_id)).one()
    instrument_name = session_entry.instrument_name

    # Update visit end time in database
    session_entry.visit_end_time = end_time
    db.add(session_entry)
    db.commit()

    # Update the multigrid controller
    data = {}
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'update_multigrid_controller_visit_end_time', session_id=session_id)}?end_time={quote(end_time.isoformat())}",
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


class ProvidedProcessingParameters(BaseModel):
    dose_per_frame: float
    extract_downscale: bool = True
    particle_diameter: Optional[float] = None
    symmetry: str = "C1"
    eer_fractionation: int = 20


@router.post("/sessions/{session_id}/provided_processing_parameters")
async def pass_proc_params_to_instrument_server(
    session_id: MurfeySessionID, proc_params: ProvidedProcessingParameters, db=murfey_db
):
    session = db.exec(select(Session).where(Session.id == session_id)).one()

    session_processing_parameters = SessionProcessingParameters(
        session_id=session_id,
        dose_per_frame=proc_params.dose_per_frame,
        gain_ref=session.current_gain_ref,
        symmetry=proc_params.symmetry,
        eer_fractionation=proc_params.eer_fractionation,
    )
    db.add(session_processing_parameters)
    db.commit()

    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        label = db.exec(select(Session).where(Session.id == session_id)).one().name
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'register_processing_parameters', session_id=session_id)}",
                json={
                    "label": label,
                    "params": {
                        "dose_per_frame": proc_params.dose_per_frame,
                        "symmetry": proc_params.symmetry,
                        "eer_fractionation": proc_params.eer_fractionation,
                        "gain_ref": session.current_gain_ref,
                    },
                },
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


@router.get("/instruments/{instrument_name}/instrument_server")
async def check_instrument_server(instrument_name: MurfeyInstrumentName):
    data = None
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.get(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'health')}",
            ) as resp:
                data = await resp.json()
    return data


@router.get(
    "/instruments/{instrument_name}/sessions/{session_id}/possible_gain_references"
)
async def get_possible_gain_references(
    instrument_name: MurfeyInstrumentName, session_id: MurfeySessionID
) -> List[File]:
    data = []
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with lock:
            token = instrument_server_tokens[session_id]["access_token"]
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.get(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'get_possible_gain_references', instrument_name=sanitise(instrument_name), session_id=session_id)}",
                headers={"Authorization": f"Bearer {token}"},
            ) as resp:
                data = await resp.json()
    return data


class GainReferenceRequest(BaseModel):
    gain_path: Path


@router.post("/sessions/{session_id}/upload_gain_reference")
async def request_gain_reference_upload(
    session_id: MurfeySessionID,
    gain_reference_request: GainReferenceRequest,
    db=murfey_db,
):
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    visit = db.exec(select(Session).where(Session.id == session_id)).one().visit
    visit_path = f"{datetime.datetime.now().year}/{visit}"
    data = {}
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'upload_gain_reference', instrument_name=instrument_name, session_id=session_id)}",
                json={
                    "gain_path": str(gain_reference_request.gain_path),
                    "visit_path": visit_path,
                    "gain_destination_dir": machine_config.gain_directory_name,
                },
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


@router.post("/visits/{visit_name}/{session_id}/upstream_tiff_data_request")
async def request_upstream_tiff_data_download(
    visit_name: str, session_id: MurfeySessionID, db=murfey_db
):
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.upstream_data_download_directory:
        download_dir = str(
            machine_config.upstream_data_download_directory
            / secure_filename(visit_name)
        )
        if machine_config.instrument_server_url:
            async with aiohttp.ClientSession() as clientsession:
                async with clientsession.post(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'gather_upstream_tiffs', visit_name=secure_filename(visit_name), session_id=session_id)}",
                    json={"download_dir": download_dir},
                    headers={
                        "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                    },
                ) as resp:
                    data = await resp.json()
    return data


class RsyncerSource(BaseModel):
    source: str


@router.post("/sessions/{session_id}/stop_rsyncer")
async def stop_rsyncer(
    session_id: MurfeySessionID, rsyncer_source: RsyncerSource, db=murfey_db
):
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'stop_rsyncer', session_id=session_id)}",
                json={
                    "source": str(secure_path(Path(rsyncer_source.source))),
                },
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


@router.post("/sessions/{session_id}/finalise_rsyncer")
async def finalise_rsyncer(
    session_id: MurfeySessionID, rsyncer_source: RsyncerSource, db=murfey_db
):
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'finalise_rsyncer', session_id=session_id)}",
                json={
                    "source": str(secure_path(Path(rsyncer_source.source))),
                },
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


@router.post("/sessions/{session_id}/finalise_session")
async def finalise_session(session_id: MurfeySessionID, db=murfey_db):
    log.debug(f"Finalising session {session_id}")
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api.router', 'finalise_session', session_id=session_id)}",
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
                log.debug(f"Received response {data}")
    return data


@router.post("/sessions/{session_id}/abandon_session")
async def abandon_session(session_id: MurfeySessionID, db=murfey_db):
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if machine_config.instrument_server_url:
        async with aiohttp.ClientSession() as clientsession:
            async with clientsession.post(
                f"{machine_config.instrument_server_url}{url_path_for('api_router', 'abandon_controller', session_id=session_id)}",
                headers={
                    "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                },
            ) as resp:
                data = await resp.json()
    return data


@router.post("/sessions/{session_id}/remove_rsyncer")
async def remove_rsyncer(
    session_id: MurfeySessionID, rsyncer_source: RsyncerSource, db=murfey_db
):
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if isinstance(session_id, int):
        if machine_config.instrument_server_url:
            async with aiohttp.ClientSession() as clientsession:
                async with clientsession.post(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'remove_rsyncer', session_id=session_id)}",
                    json={
                        "source": str(secure_path(Path(rsyncer_source.source))),
                    },
                    headers={
                        "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                    },
                ) as resp:
                    data = await resp.json()
    return data


@router.post("/sessions/{session_id}/restart_rsyncer")
async def restart_rsyncer(
    session_id: MurfeySessionID, rsyncer_source: RsyncerSource, db=murfey_db
):
    data = {}
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if isinstance(session_id, int):
        if machine_config.instrument_server_url:
            async with aiohttp.ClientSession() as clientsession:
                async with clientsession.post(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'restart_rsyncer', session_id=session_id)}",
                    json={
                        "source": str(secure_path(Path(rsyncer_source.source))),
                    },
                    headers={
                        "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                    },
                ) as resp:
                    data = await resp.json()
    return data


@router.post("/sessions/{session_id}/flush_skipped_rsyncer")
async def flush_skipped_rsyncer(
    session_id: MurfeySessionID, rsyncer_source: RsyncerSource, db=murfey_db
):
    # Load data for session
    session_entry = db.exec(select(Session).where(Session.id == session_id)).one()
    instrument_name = session_entry.instrument_name

    # Define a new visit end time that's slightly ahead of current time
    new_end_time = datetime.datetime.now().replace(
        second=0, microsecond=0
    ) + datetime.timedelta(minutes=5)
    # Update the stored visit end time if the new one exceeds it
    if session_entry.visit_end_time:
        if new_end_time > session_entry.visit_end_time:
            session_entry.visit_end_time = new_end_time
            db.add(session_entry)
            db.commit()

    # Send request to flush rsyncer
    data: dict = {}
    update_result: dict = {}
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    if isinstance(session_id, int):
        if machine_config.instrument_server_url:
            async with aiohttp.ClientSession() as clientsession:
                # Send request to instrument server to update multigrid controller
                async with clientsession.post(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'update_multigrid_controller_visit_end_time', session_id=session_id)}?end_time={quote(session_entry.visit_end_time.isoformat())}",
                    headers={
                        "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                    },
                ) as resp:
                    update_result = await resp.json()
                if not update_result.get("success", False):
                    return {"success": False}
                # Send request to flush the rsyncer
                async with clientsession.post(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'flush_skipped_rsyncer', session_id=session_id)}",
                    json={
                        "source": str(secure_path(Path(rsyncer_source.source))),
                    },
                    headers={
                        "Authorization": f"Bearer {instrument_server_tokens[session_id]['access_token']}"
                    },
                ) as resp:
                    data = await resp.json()
    return data


class RSyncerInfo(BaseModel):
    source: str
    num_files_transferred: int
    num_files_in_queue: int
    num_files_to_analyse: int
    alive: bool
    stopping: bool
    analyser_alive: bool
    analyser_stopping: bool
    destination: str
    tag: str
    files_transferred: int
    files_counted: int
    transferring: bool
    session_id: int
    num_files_skipped: int = 0


@router.get("/instruments/{instrument_name}/sessions/{session_id}/rsyncer_info")
async def get_rsyncer_info(
    instrument_name: MurfeyInstrumentName, session_id: MurfeySessionID, db=murfey_db
) -> List[RSyncerInfo]:
    rsyncer_list = []
    analyser_list = []
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    rsync_instances = db.exec(
        select(RsyncInstance).where(RsyncInstance.session_id == session_id)
    ).all()
    if machine_config.instrument_server_url:
        try:
            async with lock:
                token = instrument_server_tokens[session_id]["access_token"]
            async with aiohttp.ClientSession() as clientsession:
                async with clientsession.get(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'get_rsyncer_info', session_id=session_id)}",
                    headers={"Authorization": f"Bearer {token}"},
                ) as resp:
                    if resp.status == 200:
                        rsyncer_list = await resp.json()
                    else:
                        rsyncer_list = []
        except KeyError:
            rsyncer_list = []
        except Exception:
            log.warning(
                "Exception encountered gathering rsyncer info from the instrument server",
                exc_info=True,
            )

        try:
            async with lock:
                token = instrument_server_tokens[session_id]["access_token"]
            async with aiohttp.ClientSession() as clientsession:
                async with clientsession.get(
                    f"{machine_config.instrument_server_url}{url_path_for('api.router', 'get_analyser_info', session_id=session_id)}",
                    headers={"Authorization": f"Bearer {token}"},
                ) as resp:
                    if resp.status == 200:
                        analyser_list = await resp.json()
                    else:
                        analyser_list = []
        except KeyError:
            analyser_list = []
        except Exception:
            log.warning(
                "Exception encountered gathering analyser info from the instrument server",
                exc_info=True,
            )

    combined_data = []
    rsyncer_source_lookup = {d["source"]: d for d in rsyncer_list}
    analyser_source_lookup = {d["source"]: d for d in analyser_list}
    for ri in rsync_instances:
        rsync_data = rsyncer_source_lookup.get(ri.source, {})
        analyser_data = analyser_source_lookup.get(ri.source, {})
        combined_data.append(
            RSyncerInfo(
                source=ri.source,
                num_files_transferred=rsync_data.get("num_files_transferred", 0),
                num_files_in_queue=rsync_data.get("num_files_in_queue", 0),
                num_files_to_analyse=analyser_data.get("num_files_in_queue", 0),
                alive=rsync_data.get("alive", False),
                stopping=rsync_data.get("stopping", True),
                analyser_alive=analyser_data.get("alive", False),
                analyser_stopping=analyser_data.get("stopping", True),
                destination=ri.destination,
                tag=ri.tag,
                files_transferred=ri.files_transferred,
                files_counted=ri.files_counted,
                transferring=ri.transferring,
                session_id=session_id,
                num_files_skipped=rsync_data.get("num_files_skipped", 0),
            )
        )
    return combined_data



from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional

import requests
import xmltodict

from murfey.client.context import Context
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.util.api import url_path_for
from murfey.util.client import authorised_requests

logger = logging.getLogger("murfey.client.contexts.fib")

requests.get, requests.post, requests.put, requests.delete = authorised_requests()


class Lamella(NamedTuple):
    name: str
    number: int
    angle: Optional[float] = None


class MillingProgress(NamedTuple):
    file: Path
    timestamp: float


def _number_from_name(name: str) -> int:
    return int(
        name.strip().replace("Lamella", "").replace("(", "").replace(")", "") or 1
    )


class FIBContext(Context):
    def __init__(self, acquisition_software: str, basepath: Path):
        super().__init__("FIB", acquisition_software)
        self._basepath = basepath
        self._milling: Dict[int, List[MillingProgress]] = {}
        self._lamellae: Dict[int, Lamella] = {}

    def post_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ):
        super().post_transfer(transferred_file, environment=environment, **kwargs)
        if self._acquisition_software == "autotem":
            parts = transferred_file.parts
            if "DCImages" in parts and transferred_file.suffix == ".png":
                lamella_name = parts[parts.index("Sites") + 1]
                lamella_number = _number_from_name(lamella_name)
                time_from_name = transferred_file.name.split("-")[:6]
                timestamp = datetime.timestamp(
                    datetime(
                        year=int(time_from_name[0]),
                        month=int(time_from_name[1]),
                        day=int(time_from_name[2]),
                        hour=int(time_from_name[3]),
                        minute=int(time_from_name[4]),
                        second=int(time_from_name[5]),
                    )
                )
                if not self._lamellae.get(lamella_number):
                    self._lamellae[lamella_number] = Lamella(
                        name=lamella_name,
                        number=lamella_number,
                    )
                if not self._milling.get(lamella_number):
                    self._milling[lamella_number] = [
                        MillingProgress(
                            timestamp=timestamp,
                            file=transferred_file,
                        )
                    ]
                else:
                    self._milling[lamella_number].append(
                        MillingProgress(
                            timestamp=timestamp,
                            file=transferred_file,
                        )
                    )
                gif_list = [
                    l.file
                    for l in sorted(
                        self._milling[lamella_number], key=lambda x: x.timestamp
                    )
                ]
                if environment:
                    raw_directory = Path(
                        environment.default_destinations[self._basepath]
                    ).name
                    # post gif list to gif making API call
                    requests.post(
                        f"{environment.url.geturl()}{url_path_for('workflow.correlative_router', 'make_gif', year=datetime.now().year, visit_name=environment.visit, session_id=environment.murfey_session)}",
                        json={
                            "lamella_number": lamella_number,
                            "images": gif_list,
                            "raw_directory": raw_directory,
                        },
                    )
            elif transferred_file.name == "ProjectData.dat":
                with open(transferred_file, "r") as dat:
                    try:
                        for_parsing = dat.read()
                    except Exception:
                        logger.warning(f"Failed to parse file {transferred_file}")
                        return
                    metadata = xmltodict.parse(for_parsing)
                sites = metadata["AutoTEM"]["Project"]["Sites"]["Site"]
                for site in sites:
                    number = _number_from_name(site["Name"])
                    milling_angle = site["Workflow"]["Recipe"][0]["Activites"][
                        "MillingAngleActivity"
                    ].get("MillingAngle")
                    if self._lamellae.get(number) and milling_angle:
                        self._lamellae[number]._replace(
                            angle=float(milling_angle.split(" ")[0])
                        )



import json
import subprocess
import sys
from pathlib import Path
from queue import Empty
from unittest import mock

from murfey.cli import repost_failed_calls


@mock.patch("murfey.cli.repost_failed_calls.PikaTransport")
@mock.patch("murfey.cli.repost_failed_calls.Queue")
def test_dlq_purge(mock_queue, mock_transport, tmp_path):
    """Test the dlq purging function.
    Currently doesn't test saving the message, as the subscribe is mocked out"""
    mock_queue().get.return_value = {"message": "dummy"}
    mock_queue().get.side_effect = [None, Empty]

    exported_messages = repost_failed_calls.dlq_purge(
        tmp_path / "DLQ", "dummy", tmp_path / "config_file"
    )

    # The transport should be connected to and subscribes to the queue
    mock_transport.assert_called_once()
    mock_transport().load_configuration_file.assert_called_with(
        tmp_path / "config_file"
    )
    mock_transport().connect.assert_called_once()
    mock_transport().subscribe.assert_called_with(
        "dlq.dummy",
        mock.ANY,
        acknowledgement=True,
    )
    mock_transport().disconnect.assert_called_once()

    # Should read from the queue
    mock_queue().get.assert_any_call(True, 0.1)

    # Ideally this test would return the message, but the partial isn't called yet
    assert exported_messages == []


@mock.patch("murfey.cli.repost_failed_calls.PikaTransport")
def test_handle_dlq_messages(mock_transport, tmp_path):
    """Reinject some example messages"""
    # Create two sample messages
    messages_paths_list: list[Path] = [tmp_path / "not_a_message"]
    messages_dict: dict[str, dict] = {
        "msg1": {
            "header": {
                "x-death": [{"queue": "queue_msg1"}],
                "message-id": 1,
                "routing_key": "dlq.queue_msg1",
                "redelivered": True,
                "exchange": "",
                "consumer_tag": "1",
                "delivery_mode": 2,
                "other_key": "value",
            },
            "message": {"parameters": "msg1"},
        },
        "msg2": {
            "header": {"x-death": [{"queue": "queue_msg2"}]},
            "message": {"content": "msg2"},
        },
    }
    for file_name, message in messages_dict.items():
        messages_paths_list.append(tmp_path / file_name)
        with open(tmp_path / file_name, "w") as msg_file:
            json.dump(message, msg_file)

    # Send the two messages, plus a file that is not a message
    repost_failed_calls.handle_dlq_messages(
        messages_path=messages_paths_list,
        rabbitmq_credentials=tmp_path / "config_file",
    )

    mock_transport.assert_called_once()
    mock_transport().load_configuration_file.assert_called_with(
        tmp_path / "config_file"
    )
    mock_transport().connect.assert_called_once()

    # Only two messages should have been sent, the rest are invalid so are skipped
    assert mock_transport().send.call_count == 2
    mock_transport().send.assert_any_call(
        "queue_msg1",
        {"parameters": "msg1"},
        headers={
            "x-death": "[{'queue': 'queue_msg1'}]",
            "other_key": "value",
            "dlq-reinjected": "True",
        },
    )
    mock_transport().send.assert_any_call(
        "queue_msg2",
        {"content": "msg2"},
        headers={"x-death": "[{'queue': 'queue_msg2'}]", "dlq-reinjected": "True"},
    )

    # Removal and waiting
    assert not (tmp_path / "msg1").is_file()
    assert not (tmp_path / "msg2").is_file()
    mock_transport().disconnect.assert_called_once()


@mock.patch("murfey.cli.repost_failed_calls.requests")
def test_handle_failed_posts(mock_requests, tmp_path):
    """Test that the API is called with any failed client post messages"""
    # Create some sample messages
    messages_paths_list: list[Path] = []
    messages_dict: dict[str, dict] = {
        "msg1": {
            "message": {"url": "sample/url", "json": {"content": "msg1"}},
        },
        "msg2": {
            "message": {"url": "sample/url", "json": {"content": "msg2"}},
        },
        "msg3": {
            "message": {"content": "msg3"},  # not a failed client post
        },
        "msg4": {
            "header": {"content": "msg3"},  # does not have a message
        },
    }
    for file_name, message in messages_dict.items():
        messages_paths_list.append(tmp_path / file_name)
        with open(tmp_path / file_name, "w") as msg_file:
            json.dump(message, msg_file)

    class Response:
        def __init__(self, status_code):
            self.status_code = status_code

    mock_requests.post.side_effect = [Response(200), Response(300)]

    repost_failed_calls.handle_failed_posts(messages_paths_list, "dummy_token")

    # Check the failed posts were resent
    assert mock_requests.post.call_count == 2
    mock_requests.post.assert_any_call(
        "sample/url",
        json={"content": "msg1"},
        headers={"Authorization": "Bearer dummy_token"},
    )
    mock_requests.post.assert_any_call(
        "sample/url",
        json={"content": "msg2"},
        headers={"Authorization": "Bearer dummy_token"},
    )

    # Check only the failed post which was successfully reinjected got deleted
    assert not (tmp_path / "msg1").is_file()  # got resent
    assert (tmp_path / "msg2").is_file()  # failed reinjection
    assert (tmp_path / "msg3").is_file()  # not a failed client post
    assert (tmp_path / "msg4").is_file()  # does not have a message


@mock.patch("murfey.cli.repost_failed_calls.dlq_purge")
@mock.patch("murfey.cli.repost_failed_calls.handle_failed_posts")
@mock.patch("murfey.cli.repost_failed_calls.handle_dlq_messages")
@mock.patch("murfey.cli.repost_failed_calls.jwt")
def test_run_repost_failed_calls(
    mock_jwt,
    mock_reinject,
    mock_repost,
    mock_purge,
    mock_security_configuration,
):
    mock_jwt.encode.return_value = "dummy_token"
    mock_purge.return_value = ["/path/to/msg1"]

    config_file = mock_security_configuration
    with open(config_file) as f:
        security_config = json.load(f)

    sys.argv = [
        "murfey.repost_failed_calls",
        "--config",
        str(config_file),
        "--username",
        "user",
        "--dir",
        "DLQ_dir",
    ]
    repost_failed_calls.run()

    mock_jwt.encode.assert_called_with(
        {"user": "user"},
        security_config["auth_key"],
        algorithm=security_config["auth_algorithm"],
    )

    mock_purge.assert_called_once_with(
        Path("DLQ_dir"),
        "murfey_feedback",
        Path(security_config["rabbitmq_credentials"]),
    )
    mock_repost.assert_called_once_with(["/path/to/msg1"], "dummy_token")
    mock_reinject.assert_called_once_with(
        ["/path/to/msg1"], Path(security_config["rabbitmq_credentials"])
    )


def test_repost_failed_calls_exists():
    """Test the CLI is made"""
    result = subprocess.run(
        [
            "murfey.repost_failed_calls",
            "--help",
        ],
        capture_output=True,
    )
    assert not result.returncode

    # Find the first line of the help and strip out all the spaces and newlines
    stdout_as_string = result.stdout.decode("utf8", "replace")
    cleaned_help_line = (
        stdout_as_string.split("\n\n")[0].replace("\n", "").replace(" ", "")
    )
    assert cleaned_help_line == (
        "usage:murfey.repost_failed_calls[-h]-cCONFIG-uUSERNAME[-dDIR]"
    )



from __future__ import annotations

import argparse
import logging
import os
from threading import Thread
from typing import Literal

import graypy
import uvicorn
from rich.logging import RichHandler
from workflows import Error as WorkflowsError
from workflows.transport.pika_transport import PikaTransport

import murfey
import murfey.server
from murfey.server.feedback import feedback_listen
from murfey.server.ispyb import TransportManager
from murfey.util import LogFilter
from murfey.util.config import get_microscope, get_security_config

logger = logging.getLogger("murfey.server.run")


def _set_up_logging(quiet: bool, verbosity: int):
    rich_handler = RichHandler(enable_link_path=False)
    if quiet:
        rich_handler.setLevel(logging.INFO)
        log_levels = {
            "murfey": logging.INFO,
            "uvicorn": logging.WARNING,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.WARNING,
        }
    elif verbosity <= 0:
        rich_handler.setLevel(logging.INFO)
        log_levels = {
            "murfey": logging.DEBUG,
            "uvicorn": logging.INFO,
            "uvicorn.access": logging.WARNING,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.WARNING,
        }
    elif verbosity <= 1:
        rich_handler.setLevel(logging.DEBUG)
        log_levels = {
            "": logging.INFO,
            "murfey": logging.DEBUG,
            "uvicorn": logging.INFO,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.WARNING,
        }
    elif verbosity <= 2:
        rich_handler.setLevel(logging.DEBUG)
        log_levels = {
            "": logging.INFO,
            "murfey": logging.DEBUG,
            "uvicorn": logging.DEBUG,
            "fastapi": logging.DEBUG,
            "starlette": logging.DEBUG,
            "sqlalchemy": logging.WARNING,
        }
    else:
        rich_handler.setLevel(logging.DEBUG)
        log_levels = {
            "": logging.DEBUG,
            "murfey": logging.DEBUG,
            "uvicorn": logging.DEBUG,
            "fastapi": logging.DEBUG,
            "starlette": logging.DEBUG,
            "sqlalchemy": logging.DEBUG,
        }

    logging.getLogger().addHandler(rich_handler)
    for logger_name, log_level in log_levels.items():
        logging.getLogger(logger_name).setLevel(log_level)


def _set_up_transport(transport_type: Literal["PikaTransport"]):
    # Update the existing TransportManager object in 'murfey.server'
    murfey.server._transport_object = TransportManager(transport_type)


def run():
    """
    Main function that starts up the Murfey server
    """

    # Set up argument parser
    parser = argparse.ArgumentParser(description="Start the Murfey server")
    parser.add_argument(
        "--host",
        help="Listen for incoming connections on a specific interface (IP address or hostname; default: all)",
        default="0.0.0.0",
    )
    parser.add_argument(
        "--port",
        help="Listen for incoming TCP connections on this port (default: 8000)",
        type=int,
        default=8000,
    )
    parser.add_argument(
        "--workers", help="Number of workers for Uvicorn server", type=int, default=2
    )
    parser.add_argument(
        "--demo",
        action="store_true",
    )
    parser.add_argument(
        "--feedback",
        action="store_true",
    )
    parser.add_argument(
        "--temporary",
        action="store_true",
    )
    parser.add_argument(
        "--root-path",
        default="",
        type=str,
        help="Uvicorn root path for use in conjunction with a proxy",
    )
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        help="Decrease logging output verbosity",
    )
    verbosity.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Increase logging output verbosity",
        default=0,
    )
    # Parse and separate known and unknown args
    args, unknown = parser.parse_known_args()

    # Load the security configuration
    security_config = get_security_config()

    # Set up GrayLog handler if provided in the configuration
    if security_config.graylog_host:
        handler = graypy.GELFUDPHandler(
            security_config.graylog_host, security_config.graylog_port, level_names=True
        )
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
    # Install a log filter to all existing handlers.
    LogFilter.install()

    if args.demo:
        # Run in demo mode with no connections set up
        os.environ["MURFEY_DEMO"] = "1"
    else:
        # Load RabbitMQ configuration and set up the connection
        try:
            PikaTransport().load_configuration_file(
                security_config.rabbitmq_credentials
            )
            _set_up_transport("PikaTransport")
            logger.info("Set up message transport manager")
        except WorkflowsError:
            logger.error(
                "Error encountered setting up RabbitMQ connection",
                exc_info=True,
            )

    # Set up logging now that the desired verbosity is known
    _set_up_logging(quiet=args.quiet, verbosity=args.verbose)

    if not args.temporary and murfey.server._transport_object:
        murfey.server._transport_object.feedback_queue = security_config.feedback_queue
    rabbit_thread = Thread(
        target=feedback_listen,
        daemon=True,
    )
    logger.info("Starting Murfey RabbitMQ thread")
    if args.feedback:
        rabbit_thread.start()

    logger.info(
        f"Starting Murfey server version {murfey.__version__} for beamline {get_microscope()}, listening on {args.host}:{args.port}"
    )
    config = uvicorn.Config(
        "murfey.server.main:app",
        host=args.host,
        port=args.port,
        log_config=None,
        ws_ping_interval=300,
        ws_ping_timeout=300,
        workers=args.workers,
        root_path=args.root_path,
    )

    murfey.server._running_server = uvicorn.Server(config=config)
    murfey.server._running_server.run()
    logger.info("Server shutting down")


def shutdown():
    if murfey.server._running_server:
        murfey.server._running_server.should_exit = True
        murfey.server._running_server.force_exit = True
